#!/usr/bin/env python3

#Auteur : Pierre Koclas, May 2021
import os
import sys
import csv
from math import floor,ceil,sqrt
import matplotlib as mpl
mpl.use('Agg')
#import pylab as plt
import matplotlib.pylab as plt
import numpy as np
import matplotlib.colorbar as cbar
import matplotlib.cm as cm
import datetime
import cartopy.crs as ccrs
import cartopy.feature
#from cartopy.mpl.ticker    import LongitudeFormatter,  LatitudeFormatter
import matplotlib.colors as colors
#import matplotlib.patches as mpatches
from matplotlib.colors import ListedColormap, LinearSegmentedColormap
import sqlite3
from matplotlib.collections import PatchCollection
from statistics import median
import pikobs
import optparse
from mpl_toolkits.axes_grid1 import make_axes_locatable
def projectPpoly(PROJ,lat,lon,deltax,deltay,pc):
        X1,Y1  = PROJ.transform_point(lon - deltax,lat-deltay,pc )
        X2,Y2  = PROJ.transform_point(lon - deltax,lat+deltay,pc )
        X3,Y3  = PROJ.transform_point(lon + deltax,lat+deltay,pc )
        X4, Y4 = PROJ.transform_point(lon + deltax,lat-deltay,pc )
        Pt1=[ X1,Y1 ]
        Pt2=[ X2,Y2 ]
        Pt3=[ X3,Y3 ]
        Pt4=[ X4,Y4 ]
        Points4 = [ Pt1, Pt2,Pt3,Pt4 ]
           
        return Points4
def SURFLL(lat1,lat2,lon1,lon2):
#= (pi/180)R^2 |sin(lat1)-sin(lat2)| |lon1-lon2|
    R=6371.
    lat2=min(lat2,90.)
    surf=R*R*(np.pi/180.)*abs ( np.sin(lat2*np.pi/180.) - np.sin(lat1*np.pi/180.) ) *abs( lon2-lon1 )
   # if ( surf == 0.):
    # print (   ' surf=',lat1,lat2,lat2*np.pi/180.,lat1*np.pi/180.,np.sin(lat2*np.pi/180.) ,  np.sin(lat1*np.pi/180.) )
    return surf

def NPSURFLL(lat1, lat2, lon1, lon2):
    R = 6371.
    lat2 = np.minimum(lat2, 90.)
    surf = R**2 * (np.pi/180) * np.abs(np.sin(lat2*np.pi/180) - np.sin(lat1*np.pi/180)) * np.abs(lon2 - lon1)
  #  if np.any(surf == 0.):
    #    print('surf contiene valores cero')
    return surf
def SURFLL2(lat1, lat2, lon1, lon2):
    R = 6371.0
    lat2 = np.minimum(lat2, 90.0)
    surf = R * R * (np.pi / 180.0) * np.abs(np.sin(lat2 * np.pi / 180.0) - np.sin(lat1 * np.pi / 180.0)) * np.abs(lon2 - lon1)
    # Debugging print statements if surface is zero
    zero_surf_indices = (surf == 0.0)
 #   if np.any(zero_surf_indices):
 #       print('surf=', lat1[zero_surf_indices], lat2[zero_surf_indices], lat2[zero_surf_indices] * np.pi / 180.0,
 #             lat1[zero_surf_indices] * np.pi / 180.0,
 #             np.sin(lat2[zero_surf_indices] * np.pi / 180.0),
 #             np.sin(lat1[zero_surf_indices] * np.pi / 180.0))
    return surf
def days_between(d1, d2):
    d1 = datetime.datetime.strptime(d1, "%Y%m%d%H")
    d2 = datetime.datetime.strptime(d2, "%Y%m%d%H")
    delta = d2 - d1 

    return delta.total_seconds() / (24 * 3600)



import pikobs
import math

def round_first_digit_3(x):
    # Si es cero, lo devolvemos tal cual
    if x == 0:
        return "0"

    abs_x = abs(x)

    # Regla 1: >= 1 → 1 decimal  
    if abs_x >= 10:
        return f"{int(x)}"

    # Regla 2: >= 0.1 → 2 decimales
    elif abs_x >= 0.1:
        return f"{round(x, 2)}"

    # Regla 3: >= 0.01 → 3 decimales

    # Regla 4: >= 0.001 → 4 decimales
  #  elif abs_x >= 0.001:
   #     return f"{round(x, 3)}"

    # Regla 5: < 0.001 → formato científico
    else:
        mantissa, exp = f"{x:.1e}".split("e")
        mantissa = mantissa.split(".")[0]  # solo parte entera
        return f"{mantissa}.e{int(exp)}"

import numpy as np
def scatter_plot(
                   mode,
                   region,
                   family, 
                   id_stn, 
                   datestart,
                   dateend, 
                   Points,
                   boxsizex,
                   boxsizey, 
                   proj, 
                   pathwork, 
                   flag_criteria, 
                   fonction,
                   vcoord,
                   filesin,
                   namesin,
                   varno,
                   intervales):

       selected_flags = pikobs.flag_criteria(flag_criteria)


   
       pointsize=0.5
       delta=float(boxsizex)/2.
       deltay=float(boxsizey)/2.
       deltax=float(boxsizex)/2.
   
   #=============================================================
   #============      LECTURE   ================================
  # if isinstance(varnos, int):
  #  varnos = [varnos]
  # for fonction  in  fonctions:
  #  for Proj in proj:
       interval_a = intervales[0]
       interval_b = intervales[1]
       if  interval_a==None and  interval_b==None:
          criteria_interval = ''
          layers='layer_all'
       else:
          criteria_interval = f' and  {interval_a*100} <= vcoord <= {interval_b*100}'
          layers=f'Layer: {interval_a} hPa - {interval_b} hPa'


       conn = sqlite3.connect(":memory:")
       cursor = conn.cursor()
       cursor.execute("PRAGMA TEMP_STORE=memory")
       query = f"ATTACH DATABASE '{filesin[0]}' AS db1"  
       cursor.execute(query)

       FNAM, FNAMP, SUM, SUM2 = pikobs.type_boxes(fonction)
       if id_stn =='join' and vcoord=='join' :
          crite ="  "
       if id_stn =='join' and vcoord!='join' :
          crite = f" and  vcoord = {vcoord} "

       if id_stn !='join' and vcoord=='join' :
          crite = f"and  id_stn= '{id_stn}'    "
       
       if id_stn !='join' and vcoord!='join':
          crite = f"  and  vcoord = {vcoord} and id_stn= '{id_stn}'  "



       if len(filesin)>1:
             create_table='boites1'
             info_name = f"{namesin[0]} VS {namesin[1]}"
       else:
             create_table='AVG'
             info_name = f"namesin[0]"
        
       query = f"""CREATE TEMPORARY TABLE {create_table} AS
                   SELECT boite, 
                          lat,
                          lon, 
                          varno, 
                          vcoord,
                          SUM({SUM})/SUM(CAST(N AS FLOAT)) AVG,
                          SQRT(SUM({SUM2})/SUM(CAST(N AS FLOAT)) - SUM({SUM})/SUM(CAST(N AS FLOAT))*SUM({SUM})/SUM(CAST(N AS FLOAT))) STD,
                          SUM(sumstat)/SUM(CAST(N AS FLOAT)) BCORR,
                          SUM(n) N
                   FROM db1.moyenne
                   where varno={varno}

                   {crite} 

                   GROUP BY boite, lat, lon, varno having sum(n)>3;"""
       cursor.execute(query)

       if len(filesin)>1:
           query = f"ATTACH DATABASE '{filesin[1]}' AS db2"
           cursor.execute(query)
           query = f"""CREATE TEMPORARY TABLE boites2 AS
                       SELECT boite, lat, lon, varno, vcoord,
                              SUM({SUM})/SUM(CAST(N AS FLOAT)) AVG,

                              SQRT(SUM({SUM2})/SUM(CAST(N AS FLOAT)) - SUM({SUM})/SUM(CAST(N AS FLOAT))*SUM({SUM})/SUM(CAST(N AS FLOAT))) STD,
                              SUM(sumstat)/SUM(CAST(N AS FLOAT)) BCORR,
                              SUM(n) N
                       FROM db2.moyenne
                       where  varno={varno}

                       {crite} 
                     
                       GROUP BY boite, lat, lon, varno having sum(n)>3;"""  

           cursor.execute(query)

           query = f"""Create temporary table AVG as 
                       SELECT BOITES1.boite BOITE,
                              BOITES1.lat LAT,
                              BOITES1.lon LON,
                              BOITES1.vcoord VCOORD,
                              BOITES1.varno VARNO,
                              ABS(BOITES1.avg) - ABS(BOITES2.avg) AVG, --BOITES1.avg - BOITES2.avg AVG,
                              COALESCE(ABS(BOITES1.avg),8888) AVG1,
                              COALESCE(ABS(BOITES2.avg),8888) AVG2,
                              ABS(BOITES1.std) - ABS(BOITES2.std) STD, --  BOITES1.std - BOITES2.std STD, 
                              COALESCE(ABS(BOITES1.std),8888) as  std1,
                              COALESCE(ABS(BOITES2.std),8888) as std2,
                              ABS(BOITES1.bcorr) - ABS(BOITES2.bcorr) BCORR ,  --BOITES1.bcorr - BOITES2.bcorr BCORR , 
                              COALESCE(ABS(BOITES1.bcorr),8888) AS bcorr1,
                              COALESCE(ABS(BOITES2.bcorr),8888) AS bcorr2,
                              BOITES1.N - BOITES2.N  N , --BOITES1.N - BOITES2.N  N, 
                              COALESCE(BOITES1.N, 8888) AS N1,
                              COALESCE(BOITES2.N, 8888) AS N2
                      FROM BOITES1,BOITES2 
                      WHERE  BOITES1.boite=BOITES2.boite and BOITES1.VCOORD=BOITES2.VCOORD; 
                      """ 

           cursor.execute(query)
       
       query = f"""
        SELECT lat, lon, avg, std, N
        FROM AVG;
       """
       if len(filesin) > 1:
          query = f"""
        SELECT lat, lon, avg, std, N, N1, N2, avg1, avg2, std1, std2, bcorr, bcorr1, bcorr2
        FROM AVG;
       """

       cursor.execute(query)
      
       cursor.execute(query)
       results = cursor.fetchall()    
       # Convertir a arrays numpy
       import numpy as np
       lat = np.array([row[0] for row in results])
       lon = np.array([row[1] for row in results])
       Bomp = np.array([row[2] for row in results])
       Somp = np.array([row[3] for row in results])
       nombre = np.array([row[4] for row in results])
       
       if len(filesin) > 1:

          N1 = np.array([row[5] for row in results])
          N2 = np.array([row[6] for row in results])
          avg1 = np.array([row[7] for row in results])
          avg2 = np.array([row[8] for row in results])
          std1 = np.array([row[9] for row in results])
          std2 = np.array([row[10] for row in results])
          bcorr = np.array([row[11] for row in results])
          bcorr1 = np.array([row[12] for row in results])
          bcorr2 =  np.array([row[13] for row in results])
          con = sqlite3.connect('tu_base_de_datos.db')
          
        
          sql = """
        SELECT b1.boite, b1.lat, b1.lon
        FROM boites1 b1
        LEFT JOIN boites2 b2 ON b1.boite = b2.boite
        WHERE b2.boite IS NULL
        
        UNION
        
        SELECT b2.boite, b2.lat, b2.lon
        FROM boites2 b2
        LEFT JOIN boites1 b1 ON b2.boite = b1.boite
        WHERE b1.boite IS NULL
        """
          cursor.execute(sql)
          resultados = cursor.fetchall()
        
       dens = nombre/NPSURFLL(lat-deltay,lat+deltay,lon-deltax,lon + deltax)


       index_none=np.where(Somp ==None)
       lat = np.delete(lat, index_none) 
       lon = np.delete(lon, index_none)
       Bomp = np.delete(Bomp, index_none)
       Somp = np.delete(Somp, index_none)

       nombre = np.delete(nombre, index_none)

       query = f"""select  
         
         '{datestart}',
                  '{dateend}',
                  '{family}',
                  '{varno}' , 
                   avg(avg)  , 
                   avg(std) ,
                   sum(N) 
                   From  
                   AVG    ;"""
       
       cursor.execute(query)
       results = cursor.fetchall()   
       debut  = np.array([row[0] for row in results])
       fin    = np.array([row[1] for row in results])
       familys = np.array([row[2] for row in results])
       Mu     = np.array([row[4] for row in results])
       Sigma  = np.array([row[5] for row in results])
       Nobs   = np.array([row[6] for row in results])

       conn.close()
       typer=''
       
       import numpy as np
       import matplotlib.pyplot as plt
       from matplotlib import cm, colors
       from matplotlib.ticker import FuncFormatter
       from mpl_toolkits.axes_grid1 import make_axes_locatable
       import cartopy
       
       if Sigma is not None:
           Sigma = np.round(Sigma, 3)
       
           vartyp = fonction
           PERIODE = f'From {datestart} To {dateend}'
           NDAYS = max(1/4, days_between(datestart, dateend))
           variable_name, units, vcoord_type = pikobs.type_varno(varno)
       
           if vcoord == 'join':
               Nomvar = f"{variable_name} {units} \n id_stn:{id_stn} vcoord/channel:{(vcoord)} {layers}"
           else:
               Nomvar = f"{variable_name} {units} \n id_stn:{id_stn} vcoord/channel:{int(vcoord)} {layers} "
       
           mode = 'MOYENNE'
           OMP = Somp if mode == 'SIGMA' else Bomp
           OMP = np.nan_to_num(OMP, nan=np.nan)
       
           plt.close('all')
           fig = plt.figure(figsize=(10, 10))
           plt.rcParams['axes.linewidth'] = 1
           fontsize = 17
       
           OMPm = [value for value in OMP if isinstance(value, float)]
           vmin = np.nanmin(OMPm)
           vmax = np.nanmax(OMPm)
           Ninterv = 10
       
           if vartyp == 'dens' or vartyp == 'dens%' :
               OMP = dens / NDAYS 
               if len(filesin) > 1:
                  den1 = N1/NPSURFLL(lat-deltay,lat+deltay,lon-deltax,lon + deltax) 
                  den2 = N2/NPSURFLL(lat-deltay,lat+deltay,lon-deltax,lon + deltax) 
                  OMP = (den2-den1)*100/den1  

                  vmin = -100
                  vmax =  100
               else:
                  
             
                     vmin = np.min(OMP)
                     vmax = np.max(OMP)
               cmap_base = cm.get_cmap('PuRd', lut=Ninterv)
               if vmin < 0:
                   max_abs = max(abs(vmin), abs(vmax))
                   vmin, vmax = -max_abs, max_abs
           elif vartyp in ['nobs', 'NOBSHDR']:
               Ninterv = 10
               OMP = nombre / NDAYS
               if len(filesin) > 1:
                   N1=N1 / NDAYS
#
                   N2=N2/ NDAYS

                   OMP =(N2-N1)*100/N1
                   vmin = -100
                   vma = 100
               else:
                  vmin = np.min(OMP)
                  vmax = np.max(OMP)

               max_abs = max(abs(vmin), abs(vmax))
               if vmin > 0.0:
                   Ninterv = 11
                   cmap_base = cm.get_cmap('PuRd', lut=Ninterv)
                   vmin, vmax = vmin + 1, vmax
               else:
                   cmap_base = cm.get_cmap('RdYlBu_r', lut=Ninterv)
                   vmin, vmax = -max_abs, max_abs
           elif vartyp == 'obs':
               if len(filesin) > 1:
                   OMP =(abs(avg2)-abs(avg1))*100/abs(avg1)
                   vmin = -100
                   vmax= 100

               else:
                  vmin = np.min(OMP)
                  vmax = np.max(OMP)
                  cmap_base = cm.get_cmap('RdYlBu_r', lut=Ninterv)  
           elif vartyp == 'bcorr':
               if len(filesin) > 1:
                   OMP =(abs(bcorr2)-abs(bcorr1))*100/abs(bcorr1)
                   vmin = -100
                   vmax= 100

               else:
                  vmin = np.min(OMP)
                  vmax = np.max(OMP)
                  cmap_base = cm.get_cmap('RdYlBu_r', lut=Ninterv)

           elif vartyp in ['omp', 'oma', 'bcorr', 'stdomp']:
               vmin = np.min(OMP)
               vmax = np.max(OMP)

               if len(filesin) > 1:
                   
                   OMP =(abs(avg2)-abs(avg1))*100/abs(avg1)
                   vmin = -100
                   vmax = 100


               N = 10
               if vartyp == 'stdomp':
                   mode = 'SIGMA'
                   typer = 'STD'
               cmap_base = cm.get_cmap('seismic', lut=N)
               if mode == 'MOYENNE':
                   typer = 'AVG'
               if mode == 'SIGMA':
                   cmap_base = cm.get_cmap('RdYlBu_r', lut=Ninterv)
           else:
               cmap_base = cm.get_cmap('RdYlBu_r', lut=Ninterv)
       
           if len(filesin) > 1:
              N = 21  
              boundaries = np.linspace(vmin, vmax, N)
              y = (boundaries[:-1] + boundaries[1:]) / 2
              neg_colors = [cm.RdYlBu_r(x) for x in np.linspace(0.08, 0.40, N//2)]
              blanco = (1, 1, 1, 1) 
              blanco = (1.0, 1.0, 0.6, 1.0) 
              pos_colors = [cm.RdYlBu_r(x) for x in np.linspace(0.60, 0.92, N//2)]
              over_color = '#e600c7'   
              under_color = '#0235ad' 
              colores =  neg_colors + [blanco] + pos_colors  

              cmap = colors.ListedColormap(colores)
              cmap.set_under('#0235ad')
              cmap.set_over('#e600c7')
              norm = colors.BoundaryNorm(boundaries, N)
              extend_color = 'both'
              ticks = boundaries
           else:
               boundaries = np.linspace(vmin, vmax, Ninterv + 1)
               y = boundaries[:-1] + (boundaries[1] - boundaries[0]) / 2
               cmap = cmap_base
               norm = cm.colors.Normalize(vmin=vmin, vmax=vmax)
               extend_color = None
               ticks = boundaries
           m = cm.ScalarMappable(norm=norm, cmap=cmap)
           Colors = [m.to_rgba(x) for x in y]
           inds = np.digitize(OMP, boundaries) - 1
           inds = np.clip(inds, 0, len(Colors) - 1)
       
           nombres = 0
           left, bottom = 0.90, 0.15
           ax, fig, LATPOS, PROJ, pc = pikobs.type_projection(proj)
           ONMAP = 0 
           ONMAP2 = 0
           POINTS = 'OFF'
           patch_list = []

           for i in range(len(nombre)):
               x1, y1 = PROJ.transform_point(lon[i], lat[i], pc)
               point = PROJ.transform_point(lon[i], lat[i], src_crs=pc)
               fig_coords = ax.transData.transform(point)
               ax_coords = ax.transAxes.inverted().transform(fig_coords)
               xx, yy = ax_coords
               mask = (xx >= -0.01) & (xx <= 1.01) & (yy >= -0.01) & (yy <= 1.01)
               if mask:
                   ONMAP += nombre[i]
                   if len(filesin) > 1:
 
                       ONMAP2 +=N1[i]
                   if POINTS == 'ON':
                       plt.text(point[0], point[1], int(np.floor(nombre[i])), color="k", fontsize=17, zorder=5, ha='center', va='center', weight='bold')
                   else: 
                       if boxsizex >=10:
                          plt.text(point[0], point[1], round_first_digit_3(OMP[i]), color="b", fontsize=5, zorder=5, ha='center', va='center', weight='bold')
                     
                       points4 = projectPpoly(PROJ, lat[i], lon[i], deltax, deltay, pc)
                       
                       valor = OMP[i]

                       if valor < vmin:
                           col = '#0235ad'        
                       elif valor > vmax:
                           col = '#e600c7'         
                       else:
                           ind = np.digitize(valor, boundaries) - 1
                           col = Colors[inds[i]]     
                       if len(filesin) == 1:

                           col = Colors[inds[i]]

                       
                       poly = plt.Polygon(points4, fc=col, zorder=4, ec='k', lw=0.2, alpha=1.0)
                       ax.add_patch(poly)
           if len(filesin) > 1:
            
            for boite, latn, lonn in resultados:
             points4 = projectPpoly(PROJ, latn, lonn, deltax, deltay, pc)

             poly = plt.Polygon(points4, fc=(0,0,0,1), zorder=4, ec='k', lw=0.8, alpha=1.0)
             ax.add_patch(poly)


       
           ax.coastlines()
           ax.add_feature(cartopy.feature.LAND, zorder=1, edgecolor='#C0C0C0', facecolor='#C0C0C0')
           ax.add_feature(cartopy.feature.OCEAN, zorder=0, edgecolor='#7f7f7f', facecolor='#00bce3')
           ax.add_feature(cartopy.feature.BORDERS, zorder=10)
           ax.add_feature(cartopy.feature.COASTLINE, zorder=10)
           gl = ax.gridlines(color='b', linestyle=(0, (1, 1)), xlocs=range(-180, 190, 10), ylocs=LATPOS, draw_labels=False, zorder=0)
       
           divider = make_axes_locatable(ax)
           ax3 = divider.append_axes("right", size="5%", pad=0.05, axes_class=plt.Axes)
       
           cb2 = cbar.ColorbarBase(
               ax3, cmap=cmap,
               norm=norm,
               orientation='vertical',
               drawedges=True,
               extend= extend_color,
               ticks=ticks,
               boundaries=boundaries,
               alpha=1.0
           )
           def scientific(x, pos):
               return f'{x:.1e}'
       
           if vartyp == 'dens' :
               if len(filesin) < 1:
                    cb2.ax.yaxis.set_major_formatter(FuncFormatter(scientific))
       
           dif = ''
           if len(filesin) > 1:
               dif = 'Differences'
           if len(filesin) > 1 and   vartyp == 'dens':
                cb2.ax.set_ylabel(f'Percentage of increase in sum(dens [nobs/km²])/days \n relative to {namesin[0]} [%]', fontsize=18, rotation=90, labelpad=20)
           if len(filesin) > 1 and  vartyp == 'nobs':
                cb2.ax.set_ylabel(f'Percentage of increase in sum(nobs)/days \n relative to {namesin[0]} [%]', fontsize=18, rotation=90, labelpad=20)
           if len(filesin) > 1 and   vartyp == 'omp': 
                cb2.ax.set_ylabel(f'Percentage of improvement in OMP \n relative to {namesin[0]} [%]', fontsize=18, rotation=90, labelpad=20)
           if len(filesin) > 1 and  vartyp == 'oma' :
                cb2.ax.set_ylabel(f'Percentage of improvement in OMA \n relative to {namesin[0]} [%]', fontsize=18, rotation=90, labelpad=20)
           if len(filesin) > 1 and   vartyp == 'obs' :
                cb2.ax.set_ylabel(f'Percentage of improvement in {variable_name}  \n relative to {namesin[0]} [%]', fontsize=15, rotation=90, labelpad=20)
           if len(filesin) > 1 and   vartyp == 'bcorr' :
                cb2.ax.set_ylabel(f'Percentage increase in bcorr \n relative to {namesin[0]} [%]', fontsize=18, rotation=90, labelpad=20)
           
           if len(filesin) == 1 and   vartyp == 'dens':
                cb2.ax.set_ylabel(f'sum(dens [nobs/km²])/days', fontsize=18, rotation=90, labelpad=20)
           if len(filesin) == 1 and  vartyp == 'nobs':
                cb2.ax.set_ylabel(f'sum(nobs)/days', fontsize=18, rotation=90, labelpad=20)
           if len(filesin) == 1 and   vartyp == 'omp': 
                cb2.ax.set_ylabel(f'OMP {units}', fontsize=18, rotation=90, labelpad=20)
           if len(filesin) == 1 and  vartyp == 'oma' :
                cb2.ax.set_ylabel(f'OMA {units}', fontsize=18, rotation=90, labelpad=20)
           if len(filesin) == 1 and   vartyp == 'obs' :
                cb2.ax.set_ylabel(f'{variable_name}  {units}', fontsize=18, rotation=90, labelpad=20)
           if len(filesin) == 1 and   vartyp == 'bcorr' :
                cb2.ax.set_ylabel(f' BIAS Correction {units} ', fontsize=18, rotation=90, labelpad=20)
       
           if len(filesin) < 1:
               fontsize = 12
               ax.text(0.00, 1.05, namesin[0], fontsize=fontsize, color='b', transform=ax.transAxes) 
          
           else: 
               fontsize = 12
               start_x = 0.00
               y_coord = 1.05
               text1_len = len(namesin[0]) * 1.5
               text1_x = start_x
       
               if len(filesin) > 1:
                   text2_len = len(namesin[1])
                   text2_x = text1_x + text1_len
                   text3_x = text1_x + 0.2

                   ax.text(text1_x, y_coord, 'Diff ' + namesin[0] , fontsize=fontsize, color='blue', transform=ax.transAxes)
                   ax.text(text3_x, y_coord, "-" + namesin[1] , fontsize=fontsize, color='red', transform=ax.transAxes)
               else:
                   ax.text(text1_x, y_coord,  namesin[0] , fontsize=fontsize, color='blue', transform=ax.transAxes)
       
           ax.text(0.00 + 20, 1.05, vartyp, fontsize=fontsize, color='k', transform=ax.transAxes)
           ax.text(0.00, 1.02, PERIODE, fontsize=fontsize, color='#3366FF', transform=ax.transAxes)
           ax.text(0.45, 1.05, Nomvar, fontsize=fontsize, color='k', transform=ax.transAxes, fontweight='bold')
       
           ax.text(0.5, -0.08, 'Longitude', transform=ax.transAxes, ha='center', va='top', fontsize=14)
           ax.text(-0.06, 0.5, 'Latitude', transform=ax.transAxes, ha='center', va='bottom', rotation='vertical', fontsize=14)
       
           props = dict(boxstyle='round', facecolor='wheat', alpha=0.5)
           if vartyp in [f'dens', 'nobs']:
               textstr = f'Nobs=%.2i (for %.2f days)' % (ONMAP, NDAYS)
           else:
               textstr = '$\\bar{\\mu}=%.3f$ $\\bar{\\sigma}=%.3f$ \nNobs=%.2i (for %.2f days)' % (Mu, Sigma, ONMAP, NDAYS)
       
           if len(filesin) < 2:
               ax.text(0.82, 1.15, textstr, transform=ax.transAxes, fontsize=fontsize, verticalalignment='top', bbox=props)
           else:
               textstr = f'Diff Nobs={ONMAP:02d} \nDiff%={((ONMAP/(ONMAP2*NDAYS))*100):.2f}  % for {NDAYS:.2f} days'
 #              textstr = f'Diff Nobs=%.2i ( Diff%= %.2i in %.2f days)' % (ONMAP, (ONMAP/ONMAP2)* NDAYS*100, NDAYS)
               ax.text(0.88, 1.15, textstr, transform=ax.transAxes, fontsize=fontsize, verticalalignment='top', bbox=props)
        
           plt.grid(True)
           plt.rcParams['axes.linewidth'] = 2
       
           if vcoord == 'join':
               plt.savefig(f'{pathwork}/{family}/{fonction}_{proj}_{layers}_id_stn_{id_stn}_{region}_vcoord_{vcoord}_varno{varno}.png', dpi=600, format='png')
           else:
               plt.savefig(f'{pathwork}/{family}/{fonction}_{proj}_{layers}_id_stn_{id_stn}_{region}_vcoord_{int(vcoord)}_varno{varno}.png', dpi=600, format='png')
           plt.close(fig) 
       



#!/usr/bin/python3
###############################################################################
#
#                              plots_profils.py 
#
# Purpose: Graphs of radiance statistics 
#
# Author: Sylvain Heilliette, Pierre Koclas and David Lobon  2020
#
# Modifications:
#     -2020  David Lobon 
#        - various updates
# Syntax:
#
#    plots_profils.py 
#  
#         
###############################################################################
import sys
import csv
import math
import matplotlib as mpl
mpl.use('Agg')
import pylab
import sqlite3
import numpy as np
ROUGE = '#FF9999'
ROUGEPUR = '#FF0000'
VERT = '#009900'
BLEU = '#1569C7'
NOIR = '#000000'
COULEURS = [BLEU, ROUGEPUR, ROUGE, VERT, NOIR]


def read_sqlite(file, id_stn):
    """
    Función para leer los datos desde una base de datos SQLite y obtener los resultados de las consultas.
    """
    conn = sqlite3.connect(file)
    cursor = conn.cursor()
    query = f"""
    SELECT 
        vcoord AS vcoord,
        SUM(Ntot * AvgOMP) / SUM(Ntot) AS AvgOMP_combined,
        SUM(Ntot * AvgOMA) / SUM(Ntot) AS AvgOMA_combined,
        SQRT(SUM(Ntot * (StdOMP * StdOMP + AvgOMP * AvgOMP)) / SUM(Ntot) - POW(SUM(Ntot * AvgOMP) / SUM(Ntot), 2)) AS CombinedStdOMP,
        SQRT(SUM(Ntot * (StdOMA * StdOMA + AvgOMA * AvgOMA)) / SUM(Ntot) - POW(SUM(Ntot * AvgOMA) / SUM(Ntot), 2)) AS CombinedStdOMA,
        SUM(Ntot) AS TotalNtot,
        SUM(Ntot * AvgBCOR) / SUM(Ntot) AS AvgBCOR_combined
    FROM 
        serie_vdedr
    WHERE 
        id_stn = '{id_stn}'
    GROUP BY 
        vcoord;
    """
    cursor.execute(query)
    results = cursor.fetchall()
    conn.close()
    vcoord = [result[0] for result in results]
    AvgOMP = [result[1] for result in results]
    AvgOMA = [result[2] for result in results]
    StdOMP = [result[3] for result in results]
    StdOMA = [result[4] for result in results]
    Ntot   = [result[5] for result in results]
    BiasCor= [result[6] for result in results]
    return vcoord, AvgOMP, AvgOMA, StdOMP, StdOMA, Ntot, BiasCor

def get_graphe_nomvar():
    """
    Función que devuelve el diccionario de las variables con su descripción.
    """
    return {
        '11215': 'U COMPONENT OF WIND (10M)',
        '11216': 'V COMPONENT OF WIND (10M)',
        '12004': 'DRY BULB TEMPERATURE AT 2M',
        '10051': 'PRESSURE REDUCED TO MEAN SEA LEVEL',
        '10004': 'PRESSURE', 
        '12203': 'DEW POINT DEPRESSION (2M)',
        '12001': 'TEMPERATURE/DRY BULB', 
        '11003': 'U COMPONENT OF WIND',
        '11004': 'V COMPONENT OF WIND',
        '12192': 'DEW POINT DEPRESSION',
        '12163': 'BRIGHTNESS TEMPERATURE',
        '15036': 'ATMOSPHERIC REFRACTIVITY',
        '11001': 'WIND DIRECTION',
        '11002': 'WIND SPEED',
        '11011': 'WIND DIRECTION AT 10M', 
        '11012': 'WIND SPEED AT 10M'
    }
import matplotlib.pyplot as plt

def vdedr_plot(pathwork,datestart, dateend, flag_type, family,region, files_in,names_in ,id_stn, varno,mode):
    
    vcoord_type =  'CANAL'# LARG[1]
    files = files_in
    famille = family
    region  = region
    varno = varno
    platf = id_stn
    label = names_in[0]
    debut = datestart
    fin = dateend
    mode = mode
    #print (" varno= ", varno) 
    GRAPHE_NOMVAR = get_graphe_nomvar()
    if varno  in GRAPHE_NOMVAR :
        Nom = GRAPHE_NOMVAR[varno]
    else:
        Nom = varno
    #files.append(LARG[11])
    label2 = names_in[1]
    
    PERIODE = 'From  ' + debut + '   to  ' + fin
    
    #
    #==============================================================================
    order = True
    if ( vcoord_type == 'PRESSION'):
        SIGN = -1
        vcoord_type_e = 'Pressure'
    elif ( vcoord_type == 'HAUTEUR'):
        SIGN = -1
        vcoord_type_e = 'Height'
    elif ( vcoord_type == 'CANAL'):
        SIGN = -1
        vcoord_type_e = 'Channel'
    else:
        SIGN= -1
        vcoord_type_e = vcoord_type
    
    
    #=================================
    #fig = pylab.figure(figsize=(8, 10))
    #=================================
    #ax = fig.add_subplot(1, 1, 1)
    #ax.grid(True)
    
    filenumb = 0
    TITRE = 'VERIFS'
   # lvl = []
   # numb = []
    
    lvl1, AvgOMP1, AvgOMA1, StdOMP1, StdOMA1, Ntot1, BiasCor1 = read_sqlite(files[0], id_stn)
    lvl2, AvgOMP2, AvgOMA2, StdOMP2, StdOMA2, Ntot2, BiasCor2 = read_sqlite(files[1], id_stn)
    for mode in ['bias',"delta_err","delta_err_oma"]: 
        fig = pylab.figure(figsize=(8, 10))
          #=================================
        ax = fig.add_subplot(1, 1, 1)
        ax.grid(True)
    

        lvl = []
        numb = []
          #lvl1, Bomp1, Somp1, Nomb1 = read_delta_err_oma_csv(files[0])
         # lvl2, Bomp2, Somp2, Nomb2 = read_delta_err_oma_csv(files[1])
        lset1 = set( lvl1 )
        lset2 = set( lvl2 )
        
        common_levels = lset1 & lset2
        
        lvlm = sorted(list(common_levels), reverse = order)
        if (mode=="bias"):
          delta_residual = [] 
          delta_raw = []
        if (mode=="delta_err"):
          delta_sigma = [] 
          delta_N = []
        if (mode=="delta_err_oma"):
          delta_sigma = [] 
          delta_N = []
        
        n1 = []
        n2 = []
        idlev = range(0,len(lvlm))
        
        for lev in lvlm:
            pos1 = lvl1.index(lev)
            pos2 = lvl2.index(lev)
            if (mode=="bias"):
              delta_residual.append( AvgOMP2[pos2] - AvgOMP1[pos1]  )
              delta_raw.append(  AvgOMP2[pos2] -  BiasCor2[pos2] - AvgOMP1[pos1] + BiasCor1[pos1])
            if (mode=="delta_err"):
              delta_sigma.append( 100.0 * ( StdOMP2[pos2] -  StdOMP1[pos1]) / StdOMP1[pos1] )
              delta_N.append( 100.0 * (Ntot2[pos2] - Ntot1[pos1]) / Ntot1[pos1] )
            if (mode=="delta_err_oma"):  
              delta_sigma.append( 100.0 * (StdOMA2[pos2] -  StdOMA1[pos1]) / StdOMA1[pos1] )
              delta_N.append( 100.0 * (Ntot2[pos2] - Ntot1[pos1]) / Ntot1[pos1] )
          
            n1.append(Ntot1[pos1])
            n2.append(Ntot2[pos2])
        #========================GRAPHIQUE==============================================
        if (mode=="bias"):
          ax.plot(delta_residual, idlev, linestyle = '-', marker = 'o', \
                    color = COULEURS[2], markersize = 4 )
          ax.plot(delta_raw, idlev, linestyle = '-', marker = 'p', \
                    color = COULEURS[3], markersize = 4 ) 
        if (mode=="delta_err"):
          ax.plot(delta_sigma, idlev, linestyle = '-', marker = 'o', \
                    color = COULEURS[2], markersize = 4 )
          ax.plot(delta_N, idlev, linestyle = '-', marker = 'p', \
                    color = COULEURS[3], markersize = 4 ) 
        if (mode=="delta_err_oma"):
          ax.plot(delta_sigma, idlev, linestyle = '-', marker = 'o', \
                    color = COULEURS[2], markersize = 4 )
          ax.plot(delta_N, idlev, linestyle = '-', marker = 'p', \
                    color = COULEURS[3], markersize = 4 ) 
        
        #===============================================================================
        
        #======TICK MARKS=ET LABEL===============================
        xlim = pylab.get(pylab.gca(), 'xlim')
        if varno  in GRAPHE_NOMVAR :
            Nom2 = GRAPHE_NOMVAR[varno]
        else:
            Nom2 = varno
        ylim = (min(idlev), max(idlev) )
            
        yticks = map(str, idlev)
        ax.set_yticks(idlev)
        
        yticks = map(str, lvlm)
        ax.set_yticklabels(yticks, fontsize = 6)
        
         
        pylab.setp(pylab.gca(), ylim = ylim[::SIGN])
                
        ax.set_ylabel(vcoord_type_e, color = NOIR, bbox = dict(facecolor=ROUGE), \
                          fontsize = 16)
        #========================================================
        
        #=NOMBRE DE NONNEES ==================================================
        datapt = []
        for y in  range(0, len(lvlm) ):
            datapt.append(( xlim[1], idlev[y] ) )
        display_to_ax = ax.transAxes.inverted().transform
        data_to_display = ax.transData.transform
        
        if ( len(datapt) > 0):
            ax_pts = display_to_ax(data_to_display(datapt))
            for y in  range(0, len(lvlm) ):
                ix, iy = ax_pts[y]
                pylab.text(ix + .01 , iy, int(n1[y]), fontsize = 6, \
                               color = COULEURS[0], transform = ax.transAxes )
                pylab.text(ix + .07 , iy,int(n2[y]), fontsize = 6, \
                               color = COULEURS[1], transform = ax.transAxes )
        #====================================================================
        #----------------------------------------------------------------------

        famille1 = famille + " " + platf
        REGION1 = region
        LABEL1 = label
        LABEL2 = label2
        if (mode=="bias"):
          legendlist = [' residual bias ', ' raw bias ']
          name = "omp_bias"
          typer='OMP'
        if (mode=="delta_err"):
          legendlist = ['% sigma ', '% Nobs ']
          name = "omp_rel"
          typer='OMP'

        if (mode=="delta_err_oma"):
          legendlist = ['% sigma ', '% Nobs ']
          name = "oma_rel"
          typer='OMA'

        l1 = pylab.legend(legendlist, columnspacing=1, fancybox=False, ncol=2, \
                              shadow = False, loc = (0.50, +1.00), prop = {'size':8}, frameon=False)
        ltext = pylab.gca().get_legend().get_texts()
     #   print(ltext)
        pylab.setp(ltext[0], fontsize = 10, color = 'k')
        #----------------------------------------------------------------------
        bbox_props = dict(facecolor = BLEU, boxstyle = 'round')
        pylab.text(-.03, -0.05, f"{famille1}  {Nom} {typer} ", \
                         fontsize = 10, bbox = bbox_props, transform = ax.transAxes)
        pylab.text(.00, 1.05, REGION1, fontsize = 10, \
                        bbox = dict(facecolor = BLEU, boxstyle = 'round'), \
                        transform = ax.transAxes)
        pylab.text(.25, 1.05, PERIODE, fontsize = 10, \
                        bbox = dict(facecolor = BLEU, boxstyle = 'round'), \
                        transform = ax.transAxes)
        pylab.text(.70, 1.05, LABEL2 + ' - ' + LABEL1, fontsize = 10, \
                        bbox = dict(facecolor = BLEU, boxstyle = 'round'), \
                        transform = ax.transAxes)
        
        #==========================================================
        #pylab.show()
        pylab.savefig(f'{pathwork}/{family}/{name}_{family}_{id_stn}_{names_in[0]}-{names_in[1]}_{region}.png', format = 'png', dpi = 100)
        pylab.close()



def regions(region):
    
    """

    Available regions and their boundaries:

       - **PoleNord**: Northern polar region
         - LAT1=60, LAT2=90 (Latitude range: 60°N to 90°N)
         - LON1=-180, LON2=180 (Longitude range: 180°W to 180°E)

       - **PoleSud**: Southern polar region
         - LAT1=-90, LAT2=-60 (Latitude range: 90°S to 60°S)
         - LON1=-180, LON2=180 (Longitude range: 180°W to 180°E)

       - **AmeriqueduNord**: North America
         - LAT1=25, LAT2=60 (Latitude range: 25°N to 60°N)
         - LON1=-145, LON2=-50 (Longitude range: 145°W to 50°W)

       - **OuestAmeriqueduNord**: Western North America
         - LAT1=25, LAT2=60 (Latitude range: 25°N to 60°N)
         - LON1=-145, LON2=-97.5 (Longitude range: 145°W to 97.5°W)

       - **AmeriqueDuNordPlus**: Extended North America
         - LAT1=25, LAT2=85 (Latitude range: 25°N to 85°N)
         - LON1=-170, LON2=-40 (Longitude range: 170°W to 40°W)

       - **Monde**: World
         - LAT1=-90, LAT2=90 (Latitude range: 90°S to 90°N)
         - LON1=-180, LON2=180 (Longitude range: 180°W to 180°E)

       - **Global**: Global
         - LAT1=-90, LAT2=90 (Latitude range: 90°S to 90°N)
         - LON1=-180, LON2=180 (Longitude range: 180°W to 180°E)

       - **ExtratropiquesNord**: Northern Extratropics
         - LAT1=20, LAT2=90 (Latitude range: 20°N to 90°N)
         - LON1=-180, LON2=180 (Longitude range: 180°W to 180°E)

       - **ExtratropiquesSud**: Southern Extratropics
         - LAT1=-90, LAT2=-20 (Latitude range: 90°S to 20°S)
         - LON1=-180, LON2=180 (Longitude range: 180°W to 180°E)

       - **HemisphereNord**: Northern Hemisphere
         - LAT1=0, LAT2=90 (Latitude range: 0°N to 90°N)
         - LON1=-180, LON2=180 (Longitude range: 180°W to 180°E)

       - **HemisphereSud**: Southern Hemisphere
         - LAT1=-90, LAT2=0 (Latitude range: 90°S to 0°N)
         - LON1=-180, LON2=180 (Longitude range: 180°W to 180°E)

       - **Asie**: Asia
         - LAT1=25, LAT2=60 (Latitude range: 25°N to 60°N)
         - LON1=65, LON2=145 (Longitude range: 65°E to 145°E)

       - **Europe**: Europe
         - LAT1=25, LAT2=70 (Latitude range: 25°N to 70°N)
         - LON1=-10, LON2=28 (Longitude range: 10°W to 28°E)

       - **Mexique**: Mexico
         - LAT1=15, LAT2=30 (Latitude range: 15°N to 30°N)
         - LON1=-130, LON2=-60 (Longitude range: 130°W to 60°W)

       - **Canada**: Canada
         - LAT1=45, LAT2=90 (Latitude range: 45°N to 90°N)
         - LON1=-151, LON2=-50 (Longitude range: 151°W to 50°W)

       - **BaieDhudson**: Hudson Bay
         - LAT1=55, LAT2=90 (Latitude range: 55°N to 90°N)
         - LON1=-90, LON2=-60 (Longitude range: 90°W to 60°W)

       - **Arctiquecanadien**: Canadian Arctic
         - LAT1=58, LAT2=90 (Latitude range: 58°N to 90°N)
         - LON1=-141, LON2=-50 (Longitude range: 141°W to 50°W)

       - **EtatsUnis**: United States
         - LAT1=25, LAT2=45 (Latitude range: 25°N to 45°N)
         - LON1=-130, LON2=-70 (Longitude range: 130°W to 70°W)

       - **SudestEtatsUnis**: Southeastern United States
         - LAT1=25, LAT2=40 (Latitude range: 25°N to 40°N)
         - LON1=-100, LON2=-70 (Longitude range: 100°W to 70°W)

       - **EstAmeriqueduNord**: Eastern North America
         - LAT1=25, LAT2=60 (Latitude range: 25°N to 60°N)
         - LON1=-97.5, LON2=-50 (Longitude range: 97.5°W to 50°W)

       - **EstAmeriqueduNordPlus**: Extended Eastern North America
         - LAT1=25, LAT2=85 (Latitude range: 25°N to 85°N)
         - LON1=-97.5, LON2=-50 (Longitude range: 97.5°W to 50°W)

       - **OuestAmeriqueduNordPlus**: Extended Western North America
         - LAT1=25, LAT2=85 (Latitude range: 25°N to 85°N)
         - LON1=-170, LON2=-97.5 (Longitude range: 170°W to 97.5°W)

       - **Tropiques30**: Tropics 30
         - LAT1=-30, LAT2=30 (Latitude range: 30°S to 30°N)
         - LON1=-180, LON2=180 (Longitude range: 180°W to 180°E)

       - **Tropiques**: Tropics
         - LAT1=-20, LAT2=20 (Latitude range: 20°S to 20°N)
         - LON1=-180, LON2=180 (Longitude range: 180°W to 180°E)

       - **Australie**: Australia
         - LAT1=-55, LAT2=-10 (Latitude range: 55°S to 10°S)
         - LON1=90, LON2=180 (Longitude range: 90°E to 180°E)

       - **Pacifique**: Pacific
         - LAT1=20, LAT2=65 (Latitude range: 20°N to 65°N)
         - LON1=130, LON2=-150 (Longitude range: 130°E to 150°W)

       - **Atlantique**: Atlantic
         - LAT1=20, LAT2=65 (Latitude range: 20°N to 65°N)
         - LON1=-80, LON2=-1 (Longitude range: 80°W to 1°W)

       - **Alaska**: Alaska
         - LAT1=50, LAT2=75 (Latitude range: 50°N to 75°N)
         - LON1=-180, LON2=-140 (Longitude range: 180°W to 140°W)

       - **HIMAPEst**: HIMAP East
         - LAT1=35, LAT2=65 (Latitude range: 35°N to 65°N)
         - LON1=-105, LON2=-50 (Longitude range: 105°W to 50°W)

       - **HIMAPOuest**: HIMAP West
         - LAT1=40, LAT2=65 (Latitude range: 40°N to 65°N)
         - LON1=-135, LON2=-90 (Longitude range: 135°W to 90°W)

       - **ExtremeSud**: Extreme South
         - LAT1=-90, LAT2=-87 (Latitude range: 90°S to 87°S)
         - LON1=-180, LON2=180 (Longitude range: 180°W to 180°E)

       - **ExtremeNord**: Extreme North
         - LAT1=87, LAT2=90 (Latitude range: 87°N to 90°N)
         - LON1=-180, LON2=180 (Longitude range: 180°W to 180°E)

       - **TropiquesOuest**: Western Tropics
         - LAT1=-20, LAT2=0 (Latitude range: 20°S to 0°N)
         - LON1=180, LON2=-90 (Longitude range: 180°E to 90°W)

       - **Bande60a90**: Band from 60°N to 90°N
         - LAT1=60, LAT2=90 (Latitude range: 60°N to 90°N)
         - LON1=-180, LON2=180 (Longitude range: 180°W to 180°E)

       - **Bande30a60**: Band from 30°N to 60°N
         - LAT1=30, LAT2=60 (Latitude range: 30°N to 60°N)
         - LON1=-180, LON2=180 (Longitude range: 180°W to 180°E)

       - **Bande00a30**: Band from 0°N to 30°N
         - LAT1=0, LAT2=30 (Latitude range: 0°N to 30°N)
         - LON1=-180, LON2=180 (Longitude range: 180°W to 180°E)

       - **BandeM30a00**: Band from -30°S to 0°N
         - LAT1=-30, LAT2=0 (Latitude range: 30°S to 0°N)
         - LON1=-180, LON2=180 (Longitude range: 180°W to 180°E)

       - **BandeM60aM30**: Band from -60°S to -30°S
         - LAT1=-60, LAT2=-30 (Latitude range: 60°S to 30°S)
         - LON1=-180, LON2=180 (Longitude range: 180°W to 180°E)

       - **BandeM90aM60**: Band from -90°S to -60°S
         - LAT1=-90, LAT2=-60 (Latitude range: 90°S to 60°S)
         - LON1=-180, LON2=180 (Longitude range: 180°W to 180°E)

       - **Rapidscat**: RapidScat coverage
         - LAT1=-55, LAT2=55 (Latitude range: 55°S to 55°N)
         - LON1=-180, LON2=180 (Longitude range: 180°W to 180°E)

       - **npstere**: North Polar Stereographic
         - LAT1=0, LAT2=90 (Latitude range: 0°N to 90°N)
         - LON1=-180, LON2=180 (Longitude range: 180°W to 180°E)

       - **spstere**: South Polar Stereographic
         - LAT1=-90, LAT2=0 (Latitude range: 90°S to 0°N)
         - LON1=-180, LON2=180 (Longitude range: 180°W to 180°E)

    """

    
    if region == 'PoleNord':
        latlons = (60, 90, -180, 180)
    elif region == 'PoleSud':
        latlons = (-90, -60, -180, 180)
    elif region == 'AmeriqueduNord':
        latlons = (25, 60, -145, -50)
    elif region == 'OuestAmeriqueduNord':
        latlons = (25, 60, -145, -97.5)
    elif region == 'AmeriqueDuNordPlus':
        latlons = (25, 85, -170, -40)
    elif region == 'Monde':
        latlons = (-90, 90, -180, 180)
    elif region == 'Global':
        latlons = (-90, 90, -180, 180)
    elif region == 'ExtratropiquesNord':
        latlons = (20, 90, -180, 180)
    elif region == 'ExtratropiquesSud':
        latlons = (-90, -20, -180, 180)
    elif region == 'HemisphereNord':
        latlons = (0, 90, -180, 180)
    elif region == 'HemisphereSud':
        latlons = (-90, 0, -180, 180)
    elif region == 'Asie':
        latlons = (25, 60, 65, 145)
    elif region == 'Europe':
        latlons = (25, 70, -10, 28)
    elif region == 'Mexique':
        latlons = (15, 30, -130, -60)
    elif region == 'Canada':
        latlons = (45, 90, -151, -50)
    elif region == 'BaieDhudson':
        latlons = (55, 90, -90, -60)
    elif region == 'Arctiquecanadien':
        latlons = (58, 90, -141, -50)
    elif region == 'EtatsUnis':
        latlons = (25, 45, -130, -70)
    elif region == 'SudestEtatsUnis':
        latlons = (25, 40, -100, -70)
    elif region == 'EstAmeriqueduNord':
        latlons = (25, 60, -97.5, -50)
    elif region == 'EstAmeriqueduNordPlus':
        latlons = (25, 85, -97.5, -50)
    elif region == 'OuestAmeriqueduNordPlus':
        latlons = (25, 85, -170, -97.5)
    elif region == 'Tropiques30':
        latlons = (-30, 30, -180, 180)
    elif region == 'Tropiques':
        latlons = (-20, 20, -180, 180)
    elif region == 'Australie':
        latlons = (-55, -10, 90, 180)
    elif region == 'Pacifique':
        latlons = (20, 65, 130, -150)
    elif region == 'Atlantique':
        latlons = (20, 65, -80, -1)
    elif region == 'Alaska':
        latlons = (50, 75, -180, -140)
    elif region == 'HIMAPEst':
        latlons = (35, 65, -105, -50)
    elif region == 'HIMAPOuest':
        latlons = (40, 65, -145, -100)
    elif region == 'ExtremeSud':
        latlons = (-90, -87, -180, 180)
    elif region == 'ExtremeNord':
        latlons = (87, 90, -180, 180)
    elif region == 'TropiquesOuest':
        latlons = (-20, 0, 180, -90)
    elif region == 'Bande60a90':
        latlons = (60, 90, -180, 180)
    elif region == 'Bande30a60':
        latlons = (30, 60, -180, 180)
    elif region == 'Bande00a30':
        latlons = (0, 30, -180, 180)
    elif region == 'BandeM30a00':
        latlons = (-30, 0, -180, 180)
    elif region == 'BandeM60aM30':
        latlons = (-60, -30, -180, 180)
    elif region == 'BandeM90aM60':
        latlons = (-90, -60, -180, 180)
    elif region == 'Rapidscat':
        latlons = (-55, 55, -180, 180)
    elif region == 'npstere':
        latlons = (0, 90, -180, 180)
    elif region == 'spstere':
        latlons = (-90, 0, -180, 180)
    else:
        raise ValueError(f'Región desconocida: {region}')
    
    LAT1, LAT2, LON1, LON2 = latlons
    LATLONS = f"{LAT1}_{LAT2},{LON1}_{LON2}"
    
    return latlons

def generate_latlon_criteria(LAT1, LAT2, LON1, LON2):
    """
    This function generates a filtering criteria expression based on latitude and longitude coordinates.

    Arguments:
    LAT1 (float): The starting latitude.
    LAT2 (float): The ending latitude.
    LON1 (float): The starting longitude.
    LON2 (float): The ending longitude.

    Returns:
    str: A filtering criteria expression.
    """
    relatopLAT = '<=' if LAT2 == 90 else '<'
    relatopLON = '<=' if LON2 == 180 else '<'

    if LON1 >= LON2:
        LATLONCRIT = (
            f" and lat >= {LAT1} and lat {relatopLAT} {LAT2} "
            f" and ((lon >= 0. and lon < {LON2}) or (lon >= {LON1} and {LON1} < 179.99))"
        )
    else:
        LATLONCRIT = (
            f" and lat >= {LAT1} and lat {relatopLAT} {LAT2} "
            f" and lon >= {LON1} and lon {relatopLON} {LON2}"
        )

    return LATLONCRIT


def flag_criteria(flags):

    """     
     Flag Criteria
     =============
     
     This function generates a filtering criteria expression based on flag selection.
     
     Flag Criteria for Filtering Elements
     ------------------------------------
     
     In data processing and analysis, flag criteria are essential for selecting and managing elements based on specific conditions or quality assessments. These criteria allow users to filter data effectively, focusing on elements that meet predefined standards or have undergone particular processing stages.
     
     Possible Flag Criteria
     ----------------------
     
      - **--flag_criteria all (all elements):**
       Selects all available elements without applying restrictions based on specific bits.
     
      - **--flag_criteria assimilee (assimilated elements):**
       Selects elements that have the BIT12 (4096) active, indicating that the element has influenced the analysis.
     
      - **--flag_criteria bgckalt (approved by Background Check):**
       Excludes elements that have the BIT9 (512) and BIT11 (2048) active, indicating rejections due to AO quality control and the selection process (thinning or canal), respectively.
      
      - **--flag_criteria bgckalt_qc (approved by Background Check and QC-Var):**
       Similar to "bgckalt", but also excludes elements with BIT9 (512) and BIT11 (2048) active.
     
      - **--flag_criteria monitoring (monitoring):**
       This filter excludes elements with BIT9 (512) and BIT7 (128) active, indicating rejection by AO quality control (Background Check or QC-Var) and being in reserve, respectively.
     
      - **--flag_criteria postalt (postal):**
       This filter excludes elements with BIT17 (131072), BIT9 (512), BIT11 (2048), and BIT8 (256) active, indicating rejection by QC-Var, AO quality control, the selection process (thinning or canal), and being in the blacklist, respectively.
     
     Description of Active Bits
     --------------------------
     
     - **BIT0 (1) - ADE:** Modified or generated by ADE.
     - **BIT1 (2) - ADE:** Element that exceeds a climatological extreme or fails consistency test.
     - **BIT2 (4) - ADE:** Erroneous element.
     - **BIT3 (8) - ADE:** Potentially erroneous element.
     - **BIT4 (16) - DERIV:** Dubious element.
     - **BIT5 (32) - DERIV:** Interpolated element, generated by DERIVATE.
     - **BIT6 (64) - DERIV:** Corrected by DERIVATE sequence or bias correction.
     - **BIT7 (128) - DERIV:** Reserved.
     - **BIT8 (256) - AO:** Element rejected because it is on a blacklist.
     - **BIT9 (512) - AO:** Element rejected by AO quality control (Background Check or QC-Var).
     - **BIT10 (1024) - AO:** Generated by AO.
     - **BIT11 (2048) - AO:** Rejected by a selection process (thinning or canal).
     - **BIT12 (4096) - AO:** Element that exceeds a climatological extreme or fails consistency test.
     - **BIT13 (8192) - AO:** Comparison against test field, level 1.
     - **BIT14 (16384) - AO:** Comparison against test field, level 2.
     - **BIT15 (32768) - AO:** Comparison against test field, level 3.
     - **BIT16 (65536) - AO:** Rejected by comparison against test field (Background Check).
     - **BIT17 (131072) - AO:** Rejected by QC-Var.
     - **BIT18 (262144) - DERIV:** Not used due to orography.
     - **BIT19 (524288) - DERIV:** Not used due to land-sea mask.
     - **BIT20 (1048576) - DERIV:** Aircraft position error detected by TrackQc.
     - **BIT21 (2097152) - QC:** Inconsistency detected by a CQ process.
     
     Args: 
     
       flags (str):  The selected flag option ('all', 'rejets_qc', 'rejets_bgck', 'assimilee', 'qc', 'bias_corr', 'bgckalt', 'bgckalt_qc', 'monitoring', 'postalt').
     
     Returns:
     
     
        str: A filtering criteria expression.
 
    """

    # Define bit masks for flag criteria
    BIT6_BIASCORR = 64  
    BIT7_REJ = 128  
    BIT8_BLACKLIST = 256
    BIT9_REJ = 512
    BIT11_SELCOR = 2048  
    BIT12_VUE = 4096 
    BIT16_REJBGCK = 65536
    BIT17_QCVAR = 131072

    # Generate filtering criteria based on the selected flag
    if flags == "all":
        FLAG = "flag >= 0"
    elif flags == "assimilee":
        FLAG = f"and (flag & {BIT12_VUE}) = {BIT12_VUE}"
    elif flags == "bgckalt":
        FLAG = f"and (flag & {BIT9_REJ}) = 0 and (flag & {BIT11_SELCOR}) = 0 and (flag & {BIT8_BLACKLIST}) = 0"
    elif flags == "bgckalt_qc":
        FLAG = f"and (flag & {BIT9_REJ}) = 0 and (flag & {BIT11_SELCOR}) = 0"
    elif flags == "monitoring":
        FLAG = f"and (flag & {BIT9_REJ}) = 0 and (flag & {BIT7_REJ}) = 0"
    elif flags == "postalt":
        FLAG = f"and (flag & {BIT17_QCVAR}) = 0 and (flag & {BIT9_REJ}) = 0 and (flag & {BIT11_SELCOR}) = 0 and (flag & {BIT8_BLACKLIST}) = 0"
    elif flags == "rejets_qc":
        FLAG = f"and (flag & {BIT9_REJ}) = {BIT9_REJ}"
    elif flags == "rejets_bgck":
        FLAG = f"and (flag & {BIT16_REJBGCK}) = {BIT16_REJBGCK}"
    elif flags == "bias_corr":
        FLAG = f"and (flag & {BIT6_BIASCORR}) = {BIT6_BIASCORR}"
    elif flags == "qc":
        FLAG = f"and (flag & {BIT9_REJ}) = 0"
    else:
        raise ValueError(f'Invalid flag option: {flags}')

    return FLAG
#doctest.testmod()
#import doctest

#if __name__ == '__main__': 
#    import doctest
#    doctest.testmod()
#    args = arg_call()    



import glob
import sqlite3
import os
import matplotlib.pyplot as plt
import numpy as np
from collections import defaultdict
import json
import os
import random
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
from datetime import datetime, timedelta
def load_color_mapping( mapping_file):
        """Carga el mapeo de colores desde el archivo JSON."""
        if os.path.exists(mapping_file):
            with open(mapping_file, 'r') as file:
                try:
                    mapping = json.load(file)
                    return mapping
                except json.JSONDecodeError:
                    print("Error decoding JSON from the file. Returning empty mapping.")
                    return {}
        return {}
    
def save_color_mapping(mapping,mapping_file):
        """Guarda el mapeo de colores en el archivo JSON."""
        with open(mapping_file, 'w') as file:
            json.dump(mapping, file, indent=4)
            print("Saved color mapping:", mapping)  # Verifica el contenido guardado
    
def get_color_for_flag(flag, color_mapping,mapping_file):
        """Obtiene el color para un flag. Si no existe, asigna uno nuevo y actualiza el archivo."""
        print ('----JANINA>',color_mapping)
        if flag not in color_mapping:
            print(f"Flag {flag} not found in color mapping. Assigning a new color.")  # Mensaje de advertencia
            new_color = generate_random_color()
            color_mapping[flag] = new_color
            save_color_mapping(color_mapping,mapping_file)  # Guarda solo el nuevo mapeo
        return color_mapping[flag]
    
def generate_random_color():
        """Genera un color hexadecimal aleatorio."""
        return "#{:06x}".format(random.randint(0, 0xFFFFFF))
    

def parse_date(date_str):
    """Convierte un string en formato YYYYMMDDHH a un objeto datetime."""
    return datetime.strptime(date_str, '%Y%m%d%H')

def format_date(date):
    """Convierte un objeto datetime a un string en formato YYYYMMDDHH."""
    return date.strftime('%Y%m%d%H')

def generate_date_range(datestart_str, dateend_str):
    """Genera una lista de fechas desde datestart menos 6 horas hasta dateend en formato YYYYMMDDHH."""
    # Convierte las cadenas en objetos datetime
    datestart = parse_date(datestart_str)
    dateend = parse_date(dateend_str)

    # Resta 6 horas a datestart
   # datestart -= timedelta(hours=6)

    # Genera una lista de fechas desde datestart hasta dateend
    date_list = []
    current_date = datestart
    while current_date <= dateend:
        date_list.append(format_date(current_date))
        current_date += timedelta(hours=6)  # Cambia el intervalo según sea necesario

    return date_list# Directory and file pattern
def generate_random_color():
        """Genera un color hexadecimal aleatorio."""
        return "#{:06x}".format(random.randint(0, 0xFFFFFF))
def format_avg_obs(avg_obs, num_files):
                avg_per_file = avg_obs / num_files
                if avg_per_file >= 1:
                    return f"{int(round(avg_per_file))}"
                else:
                    # Redondear a la primera cifra significativa
                    return f"{avg_per_file:.1g}"
def format_avg_percentage(avg_percentage):
                if avg_percentage >= 0.001:
                    return f"{avg_percentage:.3f}"  # 3 decimales
                else:
                    return f"{avg_percentage:.1g}"  # Primera cifra significativa
                        # Usar la función en el código

def bits_active(number):
        if number == 0:
            return [0]  # Si el número es 0, devolver [0] explícitamente
        return [bit for bit in range(24) if (number & (1 << bit)) != 0]
import sqlite3
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from collections import defaultdict
import json
import os

def bits_to_decimal(*bit_positions):
    """
    Convierte una lista de posiciones de bits activados a su valor decimal.
    bit_positions: una lista de enteros que representan las posiciones de los bits encendidos.
    """
    decimal = 0
    for bit in bit_positions:
        decimal += 2**bit  # Sumar el valor de 2 elevado a la posición del bit
    return decimal

def parse_date(date_str):
    try:
        # Assuming date format is 'YYYYMMDDHH' or 'YYYY-MM-DD HH:MM:SS'
        return pd.to_datetime(date_str, format='%Y%m%d%H', errors='coerce')
    except ValueError:
        return pd.NaT
#def bits_active(flag):
#    """
#    Devuelve una lista de los bits activos para un flag (valor decimal).
##    """
#    return [i for i in range(24) if flag & (1 << i)]
       
def has_predefined_pair(bits,  predefined_bit_pairs):
    """
    Verifies if the active bits satisfy the conditions specified in predefined_bit_pairs.
    Handles conditions before and after multiple 'or', and supports excluded bits (e.g., 'n23').
    
    :param bits: A list of active bits (e.g., [6, 12, 15]).
    :param predefined_bit_pairs: A list of tuples representing predefined bit pairs.
                                 Example: [(12, 6, 'n23', 'or', 6, 'or', 9)].
                                 'n' prefix indicates the bit must be inactive.
    :return: A list of valid bit pairs that satisfy the conditions.
    """
    valid_pairs = []  # List to store all valid pairs

    for bit_pair in predefined_bit_pairs:
        active_conditions = []  # List of conditions to check
        excluded_bits = []  # Bits that must be inactive (prefixed with 'n')
        
        condition_sets = []  # Multiple sets of conditions, split by 'or'
        current_set = []  # The current set of conditions before the next 'or'
        
        for bit in bit_pair:
            if isinstance(bit, str) and bit.startswith('n'):
                excluded_bits.append(int(bit[1:]))  # 'n23' -> bit 23 must be inactive
            elif bit == 'or':
                # When encountering 'or', save the current set and start a new one
                condition_sets.append(current_set)
                current_set = []  # Reset for the next condition set
            else:
                current_set.append(bit)  # Add bits to the current set of conditions
        
        # Append the final set of conditions if there were no more 'or' separators
        if current_set:
            condition_sets.append(current_set)
        
        # Now check each condition set (separated by 'or')
        found_valid_set = False
        for condition_set in condition_sets:
      #      # Check if all bits in the set are active and all excluded bits are inactive
            if all(bit in bits for bit in condition_set) and all(bit not in bits for bit in excluded_bits):
                found_valid_set = True
                break  # If one set is valid, no need to check further

        if found_valid_set:
            valid_pairs.append(bit_pair)

    return valid_pairs  # Return the list of valid pairs


    return valid_pairs  # Return the list of valid pairs    
def flags_plotv2(pathwork,
               datestart,
               dateend,
               family,
               plot_title,
               vcoord,
               id_stn, 
               varno,
               region):
    db_path = f'{pathwork}/{family}/output_file{region}.db'
    
    # Connect to the SQLite database
    connection = sqlite3.connect(db_path)
    cursor = connection.cursor()
    
    dates = generate_date_range(datestart, dateend)
    dates.sort()  # Ensure dates are in order
    
    # Initialize data structures for storing results
    results_by_file = defaultdict(lambda: defaultdict(int))
    total_observations_by_file = {}
    
    # Bit descriptions for flags
    bit_descriptions = {
        0: "Modified or generated by ADE", 1: "Exceeds climate extreme", 2: "Erroneous", 3: "Possibly erroneous",
        4: "Doubtful", 5: "Interpolated", 6: "Bias corrected", 7: "Rejected by satellite QC",
        8: "Unselected (blacklisted)", 9: "Rejected for background check", 10: "Generated by analysis",
        11: "Unselected channel", 12: "Assimilated", 13: "O-P rogue check failure level 1",
        14: "O-P rogue check failure level 2", 15: "O-P rogue check failure level 3",
        16: "Rejection for O-P magnitude", 17: "Rejected by QC-Var 3DVAR", 18: "Rejected over land due to higher topography",
        19: "Rejected due to land/sea mask", 20: "Aircraft TrackQC rejection",
        21: "Rejected due to transmittance above model top", 22: "QC Obs rejection", 23: "Cloud affected radiance"
    }
    
    # Fetch total observations and flag values in one go
    for date in dates:
        query_id_stn = ''
        query_vcoord = ''
        if id_stn != 'join':
            query_id_stn = f" and id_stn='{id_stn}'"
        if vcoord != 'join':
            query_vcoord = f" and vcoord='{vcoord}'"

        cursor.execute(f"""
            SELECT flag, sum(num_data_entries)
            FROM flag_observations
            WHERE varno={varno} AND date={date} {query_id_stn} {query_vcoord}
            GROUP BY flag
        """)

        flag_data = cursor.fetchall()
        


        total_observations = sum(num for _, num in flag_data)
        total_observations_by_file[date] = total_observations
        
        for flag, num_observations in flag_data:
            results_by_file[date][flag] = num_observations

    # Normalize data and calculate cumulative percentages
    normalized_data = defaultdict(lambda: defaultdict(lambda: (0, 0)))
    numberobs_data = defaultdict(lambda: defaultdict(float))
    
    for date, counts in results_by_file.items():
        total_observations = total_observations_by_file[date]
        print ('22', total_observations )
        accumulated = 0
        for flag, num_observations in sorted(counts.items(), key=lambda item: item[1], reverse=True):
            
            percentage = (num_observations / total_observations) * 100
            normalized_data[date][flag] = (accumulated, accumulated + percentage)
            numberobs_data[date][flag] = num_observations
            accumulated += percentage

    # Calculate average percentage and observations for each flag
    average_percentage_by_flag = {}
    average_obs_by_flag = {}
    total_observations_all_files = sum(total_observations_by_file.values())
    
    all_flags = set(flag for counts in results_by_file.values() for flag in counts)
    
    for flag in all_flags:
       # print (flag, sum(results_by_file[date].get(flag, 0) for date in dates))

        total_flag_observations = sum(results_by_file[date].get(flag, 0) for date in dates)
        average_percentage_by_flag[flag] = (total_flag_observations / total_observations_all_files) * 100
        average_obs_by_flag[flag] = total_flag_observations
    
    # Prepare for plotting
    fig = plt.figure(figsize=(20, 30))
    gs = fig.add_gridspec(14, 5, height_ratios=[3000] + [1000] * 13, hspace=1)
    ax1 = fig.add_subplot(gs[0, :])
    
    file_names = list(total_observations_by_file.keys())
    x = np.arange(len(file_names))
    bar_width = 1 / len(file_names)

    # Load color mapping
    with open('color_mapping.json') as f:
        color_mapping = json.load(f)
    
    x_positions = []
    
    for file_idx, date in enumerate(dates):
        sorted_flags = sorted(average_percentage_by_flag.items(), key=lambda item: item[1], reverse=True)
        sorted_flags_keys = [flag for flag, _ in sorted_flags]
        for flag_idx, flag in enumerate(sorted_flags_keys):
            try:
                start, end = normalized_data[date][flag]
            except KeyError:
                start = 99.999
                end = 100
            label = (f'Flag {flag} (Active bits: {bits_active(flag)}, '
                     rf'Average percentage $^1$ : {format_avg_percentage(average_percentage_by_flag[flag])}%, '
                     rf'Average number of observations $^2$: {format_avg_obs(average_obs_by_flag[flag], len(dates))})') \
                    if file_idx == 0 else ""
            x_position = file_idx + bar_width * np.arange(len(dates)) - 0.5
            x_positions.extend(x_position)
            print (str(flag))
            ax1.bar(x_position, end - start, bottom=start, width=bar_width,
                    label=label, color=get_color_for_flag(str(flag), color_mapping,'color_mapping.json'))
    
    ax1.set_xlim(min(x_positions), max(x_positions) + bar_width)
    ax1.set_ylim(0, 100)
    ax1.set_yticks(np.arange(0, 101, 20))
    ax1.set_xlabel('Date', fontsize=15)
    ax1.set_ylabel('Percentage of N. Obs.', fontsize=15)
    ax1.set_title(f'Normalized Distribution of Bits Presented per File of 6h\n {datestart}-{dateend} # Family:{family} # Channel:{vcoord} # Sat:{id_stn}', fontsize=15)
    ax1.set_xticks(x[::max(1, len(file_names) // 6)])
    ax1.set_xticklabels([os.path.basename(file)[0:10] for file in file_names][::max(1, len(file_names) // 6)])
    ax1.legend(loc='upper center', bbox_to_anchor=(0.15, -0.4), ncol=1, frameon=False, fontsize=8)
    ax1.tick_params(axis='x', labelsize=18)
    ax1.tick_params(axis='y', labelsize=18)
    
    explanation_text = (
        r"$^1$ Average percentage is calculated based on the total number of observations."
        r"$^2$ Average number of observations per 6-hour intervals."
    )
    fig.text(0.5, 0.5, explanation_text, ha='center', va='top', fontsize=7, linespacing=1.5)
    # Predefined pairs of bits you want to plot
    
    # Definir pares de bits predefinidos
    print (family)
    
    if family in ('cris','iasi'):
        predefined_bit_pairs = [(0,'or',2,'or',9,'n16'),(8,),(9,16),(11,7),(11,19),(11,21),(11,23),(9,'or',11),(12,)]  # Ejemplo con exclusión de bit 23
       #a predefined_bit_pairs2 = [(0,'or',2,'or',9,'n16'),(8,),(9,16,'n8'),(11,7,'n8'),(11,19,'n8'),(11,21,'n8'),(11,23,'n8'),(12,)]  # Ejemplo con exclusión de bit 23
        predefined_bit_pairs2 = [(0,'or',2,'or',9,'n16'),(8,),(9,16,'n8'),(11,7,'n8'),(11,19,'n8'),(11,21,'n8'),(11,23,'n8'), (9,'n8','or',11,'n8'),(12,)]  # Ejemplo con exclusión de bit 23
       
    if family in ('amsua_allsky', 'amsub_allsky','atms_allsky.db' ,'atms_allsky', 'mwhs2', 'mwhs2.db','ssmis' ,'ssmis.db'): 
        predefined_bit_pairs =  [(0,'or',7),(8,),(9,16,),(9,18,),(9,23,),(9,),(12,),(23,12,),(23,),(19,)]  # Ejemplo con exclusión de bit 23
        predefined_bit_pairs_assi = [(0,'or',7),(8,),(9,16),(9,18,),(9,23,),(9),(12,),(23,12,),(23,),(19,)]  # Ejemplo con exclusión de bit 23
      
    if family in ('radar'):
       predefined_bit_pairs = [(12,),(11,),(9,'n11')]  # Ejemplo con exclusión de bit 23
       #a predefined_bit_pairs2 = [(0,'or',2,'or',9,'n16'),(8,),(9,16,'n8'),(11,7,'n8'),(11,19,'n8'),(11,21,'n8'),(11,23,'n8'),(12,)]  # Ejemplo con exclusión de bit 23
       predefined_bit_pairs2 = [(12,),(11,),(9,'n11')]  # Ejemplo con exclusión de bit 23
    if family in ('csr'):
        predefined_bit_pairs = [(0,'or',2,'or',9,'n16'),(8,),(9,16),(11,7),(11,19),(11,21),(11,23),(9,'or',11),(12,)]  # Ejemplo con exclusión de bit 23
       #a predefined_bit_pairs2 = [(0,'or',2,'or',9,'n16'),(8,),(9,16,'n8'),(11,7,'n8'),(11,19,'n8'),(11,21,'n8'),(11,23,'n8'),(12,)]  # Ejemplo con exclusión de bit 23
        predefined_bit_pairs2 = [(0,'or',2,'or',9,'n16'),(8,),(9,16,'n8'),(11,7,'n8'),(11,19,'n8'),(11,21,'n8'),(11,23,'n8'), (9,'n8','or',11,'n8'),(12,)]  # Ejemplo con exclusión de bit 23
       
        
  

    # Total de observaciones en todas las banderas (flags)
    total_observations_all_files = 0
    bit_pair_observations = defaultdict(int) 
    bit_pair_observations_assi= defaultdict(int)
    total_observations_n8 = 0
    total_observations_n8v2 = 0
    total_observations_n8v3 = 0
    total_n23= 0
    total_31_n23= 0
    total_32_n23= 0

    total_31=0
    total_32=0
    TOTAL = 0  
    TOTAL_ASSI = 0 
    for date in dates:
        cursor.execute(f"""
            SELECT flag, sum(num_data_entries), vcoord
            FROM flag_observations
            WHERE varno={varno} AND date={date}  {query_id_stn} {query_vcoord}
            GROUP BY flag, vcoord
        """)

        flag_data = cursor.fetchall()
        cursor.execute(f"""
            SELECT distinct(vcoord) FROM flag_observations  where  (flag & 4096) = 4096;
  

            
        """)
        vcoord_data = cursor.fetchall()

        def bits_activos(decimal):
                # Devuelve una lista con los números de los bits activos
                return [i for i in range(decimal.bit_length()) if decimal & (1 << i) != 0]

        vcoord_data = [vcoord_data[0] for vcoord_data in  vcoord_data ] 
        for flag, num_observations, vcoord_ in flag_data: 
            bits = bits_activos(flag)

            if vcoord_ in (vcoord_data):
               TOTAL_ASSI+= num_observations
               valid_bit_pairs  = has_predefined_pair(bits,predefined_bit_pairs)
               for bit_pair in valid_bit_pairs:

                   bit_pair_observations_assi[bit_pair] += num_observations

            TOTAL+= num_observations
          #  print (flag)
           
            valid_bit_pairs  = has_predefined_pair(bits,predefined_bit_pairs)
            
            for bit_pair in valid_bit_pairs:

               bit_pair_observations[bit_pair] += num_observations

            #valid_bit_pairs2 = has_predefined_pair(bits, predefined_bit_pairs2)
            
          #  for bit_pair in valid_bit_pairs2:
          #      bit_pair_observations2[bit_pair] += num_observations
            # Procesar todos los pares de bits válidos (puede ser más de uno)
           # for bit_pair in valid_bit_pairs:
            #    bit_pair_observations[bit_pair] += num_observations  # Solo suma los pares de bits válidosa
    for pair in bit_pair_observations:
        print ( '======>' ,pair, bit_pair_observations[pair],TOTAL)
    bit_pair_percentages = {pair: (bit_pair_observations[pair] / TOTAL) * 100 for pair in bit_pair_observations}
   # print (bit_pair_percentages2 )

    bit_pair_percentages_assi = {pair: (bit_pair_observations_assi[pair], (bit_pair_observations_assi[pair] /  TOTAL_ASSI) * 100) for pair in bit_pair_observations_assi}
  #  exit()
    # Gráfico de pie para los pares de bits
    num_cols = 5
    num_rows = (len(bit_pair_percentages) + (num_cols - 3)) // (num_cols - 2) + 1
#    bit_pair_titles = {(0,'or',2,'or', 9,'n16'): "DONNÉES ERRONÉES",
#                       (8,): "CANAUX SUR LA LISTE NOIRE",
#                       (9,16,'n8') : "BACKGROUND CHECK (O-P INNOVATION ROGUE CHECK)",
#                       (9,16) : "BACKGROUND CHECK (O-P INNOVATION ROGUE CHECK)",
#                       (11,7,'n8') : "PAS ASSIMILÉS LE JOUR"                                               ,
#                       (11,7) : "PAS ASSIMILÉS LE JOUR"                                               ,
#                       (11,19,'n8'): "SENSIBILITÉ LAND / SEA-ICE"                                               , 
#                       (11,19): "SENSIBILITÉ LAND / SEA-ICE"                                               , 
#                       (11,21,'n8'): "SENSIBILITÉ OVER MODEL TOP"                                                ,
#                       (11,21): "SENSIBILITÉ OVER MODEL TOP"                                                ,
#
#
#                       (11,23,'n8'): "AFFECTÉS PAR LES NUAGES"                                             ,
#                       (11,23     ): "AFFECTÉS PAR LES NUAGES"                                             ,
#                       (9,'n8','or',11,'n8'):"REJETÉS PAR LE CONTRÔLE DE QUALITÉ DANS SON ENSEMBLE"                                   ,
#                       (9,'or',11):"REJETÉS PAR LE CONTRÔLE DE QUALITÉ DANS SON ENSEMBLE"                                   ,
#                        (12,): "ASSIMILÉS" ,
#                       (0,'n8','or',7,'n8'): "DONNÉES ERRONÉES" , 
#                       (0,'or',7): "DONNÉES ERRONÉES" ,
#
#                       (0,'or',7,'n8'): "DONNÉES ERRONÉES" ,
#                         (8,): "CANAUX SUR LA LISTE NOIRE",
#                       (9,16,'n8'): "BACKGROUND CHECK (O-P INNOVATION ROGUE CHECK)",
#                       (9,16): "BACKGROUND CHECK (O-P INNOVATION ROGUE CHECK)",
#                       (9,18,'n8'): "SENSIBILITÉ À LA TOPOGRAPHIE",
#                       (9,18): "SENSIBILITÉ À LA TOPOGRAPHIE",
#                       (9,23,'n8'): "AFFECTÉS PAR LES NUAGES PAS ASSIMILÉS ",
#                       (9,23): "AFFECTÉS PAR LES NUAGES",
#                       (23,12): "AFFECTÉS PAR LES NUAGES MAIS ASSIMILÉS",
#                       (23,): "AFFECTÉS PAR LES NUAGES",
#
#                       (9,'n8'): "REJETÉS PAR LE CONTRÔLE DE QUALITÉ DANS SON ENSEMBLE" ,
#                       (9,): "REJETÉS PAR LE CONTRÔLE DE QUALITÉ DANS SON ENSEMBLE" ,
#
#
#
#
#
#                        }    
     
    bit_pair_titles ={(0, 'or', 2, 'or', 9, 'n16'): "ERRONEOUS DATA",
                       (8,): "CHANNELS ON THE BLACKLIST",
                       (9, 16, 'n8'): "BACKGROUND CHECK (O-P INNOVATION ROGUE CHECK)",
                       (9, 16): "BACKGROUND CHECK (O-P INNOVATION ROGUE CHECK)",
                       (11, 7, 'n8'): "NOT ASSIMILATED DURING THE DAY",
                       (11, 7): "NOT ASSIMILATED DURING THE DAY",
                       (11, 19, 'n8'): "LAND / SEA-ICE SENSITIVITY",
                       (11, 19): "LAND / SEA-ICE SENSITIVITY",
                       (11, 21, 'n8'): "SENSITIVITY OVER MODEL TOP",
                       (11, 21): "SENSITIVITY OVER MODEL TOP",
                       (11, 23, 'n8'): "AFFECTED BY CLOUDS",
                       (11, 23): "AFFECTED BY CLOUDS AND BIT 11",
                       (9, 'n8', 'or', 11, 'n8'): "REJECTED BY OVERALL QUALITY CONTROL",
                       (9, 'or', 11): "REJECTED BY OVERALL QUALITY CONTROL",
                       (12,): "ASSIMILATED",
                       (0, 'n8', 'or', 7, 'n8'): "ERRONEOUS DATA",
                       (0, 'or', 7): "ERRONEOUS DATA",
                       (0, 'or', 7, 'n8'): "ERRONEOUS DATA",
                       (8,): "CHANNELS ON THE BLACKLIST",
                       (9, 16, 'n8'): "BACKGROUND CHECK (O-P INNOVATION ROGUE CHECK)",
                       (9, 16): "BACKGROUND CHECK (O-P INNOVATION ROGUE CHECK)",
                       (9, 18, 'n8'): "TOPOGRAPHY SENSITIVITY",
                       (9, 18): "TOPOGRAPHY SENSITIVITY",
                       (9, 23, 'n8'): "AFFECTED BY CLOUDS NOT ASSIMILATED",
                       (9, 23): "AFFECTED BY CLOUDS AND BIT 9",
                       (23, 12): "AFFECTED BY CLOUDS BUT ASSIMILATED",
                       (23,): "AFFECTED BY CLOUDS",
                       (9, 'n8'): "REJECTED BY OVERALL QUALITY CONTROL",
                       (9,): "REJECTED BY OVERALL QUALITY CONTROL",
                       (11,): "REJECTED BY THINNING",
                       (9,'n11'): "REJECTED BY BACKGROUND CHECK",
                       (19,): "LAND / SEA-ICE SENSITIVITY",

                       }

                       #(9,11,'n7') : "PAS ASSIMILÉS LE JOUR",
                      # PAS ASSIMILÉS LE JOUR}
  #  print (bit_pair_percentages)
    bit_pair_percentages_sorted = dict(sorted(bit_pair_percentages.items(), key=lambda x: x[1], reverse=True))
    bit_pair_percentages_sorted_assi = dict(sorted(bit_pair_percentages_assi.items(), key=lambda x: x[1][1], reverse=True))
                 
    #import matplotlib.pyplot as plt
    for i, (bit_pair, percentage) in enumerate(bit_pair_percentages_sorted.items()):
        row = i // (num_cols - 2) + 2
        col = (i % (num_cols - 2)) + 2
        ax2 = fig.add_subplot(gs[row, col])
        
        sizes = [percentage, 100 - percentage]
        significant_threshold = 1  # Adjust this value if needed
        colors = ['tab:blue', 'white']
       # print (bit_pair, percentage)
        if percentage < significant_threshold:
             # If percentage is very small, show a more meaningful label
             labels = [f'Bits {bit_pair}\n{percentage:.3g}%', f'Rest (Bits)\n{round(100 - percentage, 3)}%']
        else:
             # Standard labels when percentage is not small
             labels = [f'Bits {bit_pair}\n{round(percentage, 3)}%', f'Rest (Bits)\n{round(100 - percentage, 3)}%']

             
        textprops = {'color': 'black', 'fontweight': 'bold'}
        print ('ssss',sizes)

        ax2.pie(sizes, labels=labels, colors=colors, autopct=None, startangle=190,
                wedgeprops={'edgecolor': 'black'}, textprops=textprops, labeldistance=1.2)  
        # Dynamically create the description for all valid bits in the bit_pair
        additional_title = bit_pair_titles.get(tuple(bit_pair), r"\textbf{Default Title}")
        bit_descriptions_str = "\n ".join([bit_descriptions.get(bit, f"Unknown ({bit})") 
                                  for bit in bit_pair 
                                  if isinstance(bit, int) or (isinstance(bit, str) and not bit.startswith('n') and bit != 'or')])
       # aax2.set_title(f'\\textbf{{{additional_title}}}\n Bits {bit_pair}: \n{bit_descriptions_str}', pad=10, fontsize=10)
        #ax2.text(0.5, 1.01, f'Bits {bit_pair}:', ha='center', va='center', fontsize=10, transform=ax2.transAxes)
        #ax2.text(0.5, 0.95, bit_descriptions_str, ha='center', va='center', fontsize=10, transform=ax2.transAxes)
        ax2.set_title(f'{additional_title}\n Bits {bit_pair}: \n{bit_descriptions_str}', pad=10, fontsize=10)
        ax2.axis('equal')
    # Prepare the percentages to be written
    percentages_text = "Pourcentages tenant compte uniquement des canaux souhaités.\n"
    percentages_text = "Percentages considering only the desired channels.\n"
    if family !='radar':
        for bit_pair, (observations, percentage)  in bit_pair_percentages_sorted_assi.items():
           # Get the title or description
           if bit_pair!= (8,) and bit_pair!=(0,'or',2,'or',9,'n16') and bit_pair!=(11,23     ):
             additional_title = bit_pair_titles.get(tuple(bit_pair), "Bits " + str(bit_pair))
           #  print (total_observations_n8v2, total_observations_all_files)
            # percentages_text += f"*{additional_title}: {(percentage/(total_observations_n8 *100/ total_observations_all_files))*100:.4g}%\n" 
        #     percentages_text += f"*{additional_title}: {percentage/((total_observations_n8-total_observations_n8v3) *100/ total_observations_all_files)   }\n"
             try:
               percentages_text += f"*{additional_title}:  {(observations*100/TOTAL_ASSI):.4g}%\n"
             except:
               print ("")
    
           if bit_pair==(0,'or',2,'or',9,'n16'):
             additional_title = bit_pair_titles.get(tuple(bit_pair), "Bits " + str(bit_pair))
             try:
               percentages_text += f"*{additional_title}: {total_observations_n8v3*100/TOTAL_ASSI:.4g}%\n"
             except:
               print ("")
            #  percentages_text += f"*{additional_title}: {(percentage/(total_observations_n8v2 *100/ total_observations_all_files))*100:.4g}%\n"
           #  print (f"Q*{additional_title}: {100- total_observations_n8*100/total_observations_n8v2:.4g}%")
        # Add the text to the figure
        fig.text(0.65, 0.5,percentages_text, ha='left', va='bottom', fontsize=12)

    print (f'{pathwork}/{family}/combined_{pathwork[0:7]}_channel_{vcoord}_id_stn_{id_stn}_{varno}_plot.png')
    plt.savefig(f'{pathwork}/{family}/combined_{pathwork[0:7]}_channel_{vcoord}_id_stn_{id_stn}_{varno}_plot.png', bbox_inches='tight')



"""
Generating Flags Info
=========================================================

This module calculates the type of flag associated with each observation for different families of observations. For example, the following representations:

 .. image:: ../../../docs/source/_static/flagatms_allsky.png
      :alt: atms allsky flag Plot

 .. image:: ../../../docs/source/_static/flagcris.png
      :alt: cris  flag Plot

 .. image:: ../../../docs/source/_static/flagiasi.png
      :alt: iasi flag Plot

Parameters Explanation
^^^^^^^^^^^^^^^^^^^^^^
- **path_experience:** Path to the directory where the experience data is stored.
- **experience_name:** Name of the specific experience.
- **pathwork:** Directory path indicating the working directory for the script.
- **datestart:** Start date and time of the assimilation process (format: YYYYMMDDHH).
- **dateend:** End date and time of the assimilation process (format: YYYYMMDDHH).
- **region:** Geographic region or area of interest for the assimilation analysis. (e.g., Monde, PoleNord, PoleSud, AmeriqueduNord, OuestAmeriqueduNord, AmeriqueDuNordPlus, ExtratropiquesNord, HemisphereNord, HemisphereSud, Asie, Europe, Mexique, Canada, BaieDhudson, Arctiquecanadien, EtatsUnis, Tropiques30, Tropiques, Australie, Pacifique and Atlantique)
- **family:** Family of data assimilated (e.g., mwhs2, mwhs2_qc, to_amsua_qc, to_amsua, to_amsua_allsky, to_amsua_allsky_qc, to_amsub_qc, to_amsub, ssmis_qc, ssmis, iasi, iasi_qc, crisfsr1_qc, crisfsr2_qc, cris, atms_allsky, atms_qc and csr, csr_qc)
- **id_stn:** all or satellite IDs for which assimilation results are analyzed (e.g., METOP-1, METOP-3). Modify as needed to specify particular satellites.
- **channel:** all or specific channels to analyze. Use 'all' to analyze all available channels. Modify as needed to specify particular channels of interest.
- **plot_title:** Title for the generated cardiogram.
- **n_cpu:** Number of CPU cores to be used for parallel processing. Adjust according to your system capabilities.

Additional Note
^^^^^^^^^^^^^^^
- Modify **--channel** and **--id_stn** parameters as needed to specify particular satellites and channels of interest.
- Ensure paths (**path_experience, pathwork**) are correctly set to point to your data directories.
- The script leverages parallel processing (**--n_cpu**) to optimize performance, adjust according to your system capabilities.
- The generated cardiograms provide comprehensive visual insights into the assimilation performance metrics across selected satellites, channels, and criteria.
- Adjust the parameters (**--channel, --id_stn, etc.**) according to your specific requirements and data.


""" 
#!/usr/bin/python3
import sqlite3
import pikobs
import re
import os
from  dask.distributed import Client
import numpy as np
import sqlite3
import os
import re
import sqlite3
from  datetime import datetime, timedelta
import glob
import sqlite3
import pikobs  # Assuming this is a custom module for region processing

def create_and_insert_flag_observations(input_file, output_file, region,pathwork,family):
    # Connect to the input SQLite file
    print (input_file)
    conn_input = sqlite3.connect(input_file)
    basename = os.path.basename(input_file)
    cursor_input = conn_input.cursor()

    # Get region criteria
    LAT1, LAT2, LON1, LON2 = pikobs.regions(region)
    LATLONCRIT = pikobs.generate_latlon_criteria(LAT1, LAT2, LON1, LON2)

    # Connect to the output SQLite file with optimizations
    conn_output = sqlite3.connect(f'{pathwork}/{family}/{output_file}', uri=True, isolation_level=None, timeout=99999)
    conn_output.execute('PRAGMA synchronous = OFF')  # Turn off synchronous mode for faster writes
    conn_output.execute('PRAGMA journal_mode = MEMORY')  # Use memory journal mode for faster writes
    cursor_output = conn_output.cursor()

    # Create a table in the output database if it does not exist
    cursor_output.execute('''
        CREATE TABLE IF NOT EXISTS flag_observations (
            date TEXT,
            flag INTEGER,
            num_data_entries INTEGER,
            id_stn INTEGER,
            varno interger,
            vcoord integer
        )
    ''')

    # Query to get the required data
    query = f'''
        SELECT 
    {basename[0:10]},
    flag, 
    COUNT(*) AS num_data_entries, 
    id_stn, 
    varno,
    vcoord
    
FROM 
    data 
natural join  
    header 
    -- where {LATLONCRIT}
GROUP BY 
    date, 
    varno,
    flag, 
    id_stn,
    vcoord;
    '''

    # Execute the query and fetch results
    cursor_input.execute(query)
    results = cursor_input.fetchall()

    # Insert results into the output database in batches
    cursor_output.executemany('''
        INSERT INTO flag_observations (date, flag, num_data_entries, id_stn, varno, vcoord)
        VALUES (?, ?, ?, ?, ?,?)
    ''', results)

    # Commit and close the connections
    conn_input.close()
    conn_output.commit()
    conn_output.close()



def create_data_list_cardio(datestart1, 
                            dateend1,
                            family,
                            pathin, 
                            pathwork,
                            flag_criteria,
                            id_stn,
                            channel,
                            region_seleccionada):
    
    data_list_cardio = []

    # Convert datestart and dateend to datetime objects
    datestart = datetime.strptime(datestart1, '%Y%m%d%H')
    dateend = datetime.strptime(dateend1, '%Y%m%d%H')

    # Initialize the current_date to datestart
    current_date = datestart

    # Define a timedelta of 6 hours
    delta = timedelta(hours=6)
    FAM, VCOORD, VCOCRIT, STATB, element, VCOTYP = pikobs.family(family)

    element_array = np.array([float(x) for x in element.split(',')])
    for varno in element_array:
   #  print ("VCOORD", vcoord, element, type(element))
     # Iterate through the date range in 6-hour intervals
     while current_date <= dateend:
        # Format the current date as a string
        formatted_date = current_date.strftime('%Y%m%d%H')

        # Build the file name using the date and family
        filename = f'{formatted_date}_{family}'

        file_path_name = f'{pathin}/{filename}'
        conn = sqlite3.connect(file_path_name)
        # Create a cursor to execute SQL queries
        cursor = conn.cursor()

        #  Create a new dictionary and append it to the list
        data_dict = {'family': family,
                     'filein': f'{pathin}/{filename}',
                     'db_new': f'{pathwork}/cardio_{datestart1}_{dateend1}_{flag_criteria}_{family}.db',
                     'region': region_seleccionada,
                     'flag_criteria': flag_criteria,
                     'varno':  varno,
                     'vcoord': channel,
                     'id_stn': id_stn}
        data_list_cardio.append(data_dict)
        conn.close()

        # Update the current_date in the loop by adding 6 hours
        current_date += delta

    return data_list_cardio


def get_station_ids(cursor, id_stn):
    if id_stn[0] == 'all':
        cursor.execute("SELECT DISTINCT id_stn FROM data natural join header;")
        return [item[0] for item in cursor.fetchall()]
    else:
        
        return np.array(id_stn)

def fetch_channels_and_varno(cursor, idstn, channel):
    criter = f'WHERE id_stn = "{idstn}"'
    if channel[0] == 'all':
        query = f"SELECT DISTINCT vcoord, varno FROM data natural join header {criter} ORDER BY vcoord ASC;"
        cursor.execute(query)
        return cursor.fetchall()
    if channel[0] == 'join':
        return channel, idstn

def create_data_list_plot2(datestart1, dateend1, family, pathwork, region, id_stn, channel):
    data_list_plot = []
    import glob
    filedb = f'{pathwork}/output_file{region}.db'
    try:
        with sqlite3.connect(filedb) as conn:
            cursor = conn.cursor()
            id_stns = get_station_ids(cursor, id_stn)

            for idstn in id_stns:
                channels_varno = fetch_channels_and_varno(cursor, idstn, channel)
                for chan, varno in channels_varno:
                    data_dict_plot = {
                        'id_stn': idstn,
                        'vcoord': chan,
                        'varno': varno
                    }
                    data_list_plot.append(data_dict_plot)
    except sqlite3.Error as e:
        print(f"Database error: {e}")
    except Exception as e:
        print(f"Error: {e}")
    return data_list_plot

def create_data_list_plot(family, pathwork, region, id_stn, vcoord):
    filedb = f'{pathwork}/{family}/output_file{region}.db'
    
    conn = sqlite3.connect(filedb)  # Conectar a la base de datos
    cursor = conn.cursor()

    data_dict_plot = []

    # Paso 1: Obtener todos los valores distintos de 'varno'
    cursor.execute("SELECT DISTINCT varno FROM flag_observations")
    varnos = cursor.fetchall()

    for varno_tuple in varnos:
        varno = varno_tuple[0]

        if vcoord[0] == 'join':
            if id_stn[0] == 'join':
                # Si 'id_stn' y 'vcoord' son 'join', solo se necesita un registro por 'varno'
                data_dict_plot.append({
                    'id_stn': 'join',
                    'vcoord': 'join',
                    'varno': varno
                })

            elif id_stn[0] == 'all':
                # Obtener 'id_stn' para cada 'varno' y 'vcoord' es 'join'
                query_id_stn = """
                    SELECT DISTINCT id_stn
                    FROM flag_observations
                    WHERE varno = ?
                """
                cursor.execute(query_id_stn, (varno,))
                id_stns = cursor.fetchall()

                for id_stn_tuple in id_stns:
                    id_stn_value = id_stn_tuple[0]
                    data_dict_plot.append({
                        'id_stn': id_stn_value,
                        'vcoord': 'join',
                        'varno': varno
                    })

        elif vcoord[0] == 'all':
            if id_stn[0] == 'join':
                # Obtener 'id_stn' para cada 'varno' y 'vcoord' es 'all'
                    
                    # Obtener 'vcoord' para cada 'id_stn'
                    query_vcoord = """
                        SELECT DISTINCT vcoord
                        FROM flag_observations
                        WHERE varno = ? 
                    """
                    cursor.execute(query_vcoord, (varno,))
                    vcoords = cursor.fetchall()

                    for vcoord_tuple in vcoords:
                        vcoord_value = vcoord_tuple[0]
                        data_dict_plot.append({
                            'id_stn': 'join',
                            'vcoord': vcoord_value,
                            'varno': varno
                        })

            elif id_stn[0] == 'all':
                # Obtener 'id_stn' y 'vcoord' para cada 'varno'
                query_id_stn = """
                    SELECT DISTINCT id_stn
                    FROM flag_observations
                    WHERE varno = ?
                """
                cursor.execute(query_id_stn, (varno,))
                id_stns = cursor.fetchall()

                for id_stn_tuple in id_stns:
                    id_stn_value = id_stn_tuple[0]
                    
                    # Obtener 'vcoord' para cada 'id_stn'
                    query_vcoord = """
                        SELECT DISTINCT vcoord
                        FROM flag_observations
                        WHERE varno = ? AND id_stn = ?
                    """
                    cursor.execute(query_vcoord, (varno, id_stn_value))
                    vcoords = cursor.fetchall()

                    for vcoord_tuple in vcoords:
                        vcoord_value = vcoord_tuple[0]
                        data_dict_plot.append({
                            'id_stn': id_stn_value,
                            'vcoord': vcoord_value,
                            'varno': varno
                        })

    conn.close()
    return data_dict_plot


def filter_files_by_date(path_experience_files, family, date1, date2):
    # Lista todos los archivos que coinciden con el patrón
    print (path_experience_files, family)
    file_list = glob.glob(os.path.join(path_experience_files, f'*{family}'))

    # Convertir las fechas de los archivos y el rango a objetos datetime
    date_format = "%Y%m%d%H"
    date1_dt = datetime.strptime(date1, date_format)
    date2_dt = datetime.strptime(date2, date_format)

    filtered_files = []

    for file in file_list:
        # Extraer la parte de la fecha del nombre del archivo
        basename = os.path.basename(file)
        date_str = basename.split('_')[0]  # Esto asume que la fecha está al principio y seguida de '_'
        
        try:
            file_date = datetime.strptime(date_str, date_format)
        except ValueError:
            # Si no se puede convertir la fecha, ignorar el archivo
            continue
        
        # Verificar si la fecha del archivo está dentro del rango
        if date1_dt <= file_date <= date2_dt:
            filtered_files.append(file)
    
    return filtered_files  

def make_flags(path_experience_files,
                experience_names,      
                pathwork, 
                datestart,
                dateend,
                regions, 
                familys, 
             #   flag_criteria, 
                id_stn,
                channel,
            #    plot_type,
                plot_title,
                n_cpu):
   
   #pikobs.delete_create_folder(pathwork, family)

   for family in familys:
     for region in regions:
       pikobs.delete_create_folder(pathwork, family)

       import time
       import dask


       file_list = filter_files_by_date(path_experience_files, family, datestart, dateend)
       output_file=f"output_file{region}.db"
       try:
           os.remove(output_file)
           print(f"File {output_file} has been deleted.")
       except FileNotFoundError:
           print(f"File {output_file} not found.")
       t0= time.time()

       if n_cpu==1: 
          print (f'Number of files in serie = {len(file_list)}')

          for input_file in file_list:
             create_and_insert_flag_observations(input_file, output_file, region, pathwork,family)
       else:

          print (f'Number of files in paralle = {len(file_list)}')
          with dask.distributed.Client(processes=True, threads_per_worker=1, 
                                           n_workers=n_cpu, 
                                           silence_logs=40) as client:
            delayed_funcs = [dask.delayed(create_and_insert_flag_observations)(
                               input_file, output_file, region, pathwork,family)for input_file in file_list]

            results = dask.compute(*delayed_funcs)
       tn= time.time() 
       print ('Total time for plotting:',tn-t0 )  
       data_list_plot = create_data_list_plot(family, pathwork, region, id_stn, channel)
      # os.makedirs(f'{pathwork}/plots_{family}')
       t0 = time.time()
       if len(data_list_plot)==0:
          raise ValueError('You must specify best selectoion od id_stn and channel plot==0')
       n_cpu=1
      # print (data_list_plot)
       if n_cpu==1: 
          print (f'in Serie plots = {len(data_list_plot)}')
          print (data_list_plot)
          for  data_ in data_list_plot: 
              print ( data_['vcoord'],data_['id_stn'],   data_['varno'])


              pikobs.flags_plotv2(pathwork,
                                 datestart,
                                 dateend,
                                 family,
                                 plot_title,
                                 data_['vcoord'],
                                 data_['id_stn'], 
                                 data_['varno'],
                                 region)
       else:
          print (f'Number of plot in paralle = {len(data_list_plot)}')
          with dask.distributed.Client(processes=True, threads_per_worker=1, 
                                           n_workers=n_cpu, 
                                           silence_logs=40) as client:
            delayed_funcs = [dask.delayed(pikobs.flags_plotv2)(
                                 pathwork,
                                 datestart,
                                 dateend,
                                 family,
                                 plot_title,
                                 data_['vcoord'],
                                 data_['id_stn'], 
                                 data_['varno'],
                                 region)for data_ in data_list_plot]

            results = dask.compute(*delayed_funcs)
       tn= time.time()
       print ('Total time for plotting:',tn-t0 )  


def arg_call():
    import argparse
    import sys

    parser = argparse.ArgumentParser()
    parser.add_argument('--path_experience_files', default='undefined', type=str, help="Directory where input sqlite files are located")
    parser.add_argument('--experience_name', default='undefined', type=str, help="experience's name")
    parser.add_argument('--pathwork', default='undefined', type=str, help="Working directory")
    parser.add_argument('--datestart', default='undefined', type=str, help="Start date")
    parser.add_argument('--dateend', default='undefined', type=str, help="End date")
    parser.add_argument('--region', nargs="+", default='undefined', type=str, help="Region")
    parser.add_argument('--family', nargs="+", default='undefined', type=str, help="Family")
  #  parser.add_argument('--flags_criteria', default='undefined', type=str, help="Flags criteria")
    parser.add_argument('--id_stn',nargs="+", default='all', type=str, help="id_stn") 
    parser.add_argument('--channel',nargs="+",  default='all', type=str, help="channel") 
  #  parser.add_argument('--plot_type', default='wide', type=str, help="channel")
    parser.add_argument('--plot_title', default='plot', type=str, help="channel")
    parser.add_argument('--n_cpus', default=1, type=int, help="Number of cpus")

    args = parser.parse_args()
    print ( "Inputs in cardiogram calculation")
    print ("----------------------------------------")

    for arg in vars(args):
      
       print (f'--{arg} {getattr(args, arg)}')
    print ("----------------------------------------")
    # check if each argument is 'undefined'
    if args.path_experience_files == 'undefined':
        raise ValueError('You must specify --path_experience_files')
    if args.experience_name == 'undefined':      
        raise ValueError('You must specify -experience_name')
    if args.pathwork  == 'undefined':
        raise ValueError('You must specify --pathwork')
    if args.datestart == 'undefined':
        raise ValueError('You must specify --datestart')
    if args.dateend == 'undefined':
        raise ValueError('You must specify --dateend')
    if args.region == 'undefined':
        raise ValueError('You must specify --region')
    if args.family == 'undefined':
        raise ValueError('You must specify --family')
 #   if args.flags_criteria == 'undefined':
 #       raise ValueError('You must specify --flags_criteria')
   # if args.fonction == 'undefined':
   #     raise ValueError('You must specify --fonction')

    #Call your function with the arguments
    sys.exit(make_flags(args.path_experience_files,
                        args.experience_name,
                        args.pathwork,
                        args.datestart,
                        args.dateend,
                        args.region,
                        args.family,
                        args.id_stn,
                        args.channel,
                        args.plot_title,
                        args.n_cpus))




import os
import sqlite3
import pandas as pd
import matplotlib.pyplot as plt
from glob import glob

# Function to extract date from filename
def extract_date_from_filename(filename, family):
    basename = os.path.basename(filename)
    date_str = basename.split(family)[0][:10]  # 'yyyymmddhh'
    return pd.to_datetime(date_str, format='%Y%m%d%H')

def get_counts_by_date(path, family, start_date_str, end_date_str):
    start_date = pd.to_datetime(start_date_str, format='%Y%m%d%H')
    end_date = pd.to_datetime(end_date_str, format='%Y%m%d%H')
    
    files = glob(os.path.join(path, f'*{family}*'))
    filtered_files = []
    for file in files:
        file_date = extract_date_from_filename(file, family)
        if start_date <= file_date <= end_date:
            filtered_files.append(file)

    date_counts = {}

    for file in filtered_files:
        conn = sqlite3.connect(file)
        cursor = conn.cursor()
        cursor.execute('SELECT COUNT(*) FROM data JOIN header ON data.id_obs = header.id_obs WHERE header.id_stn LIKE "NOA%";')
        count = cursor.fetchone()[0]
        
        file_date = extract_date_from_filename(file, family)
        date_counts[file_date] = date_counts.get(file_date, 0) + count
        
        conn.close()

    return date_counts

# Paths for the families
families = [
    ('postalt', '/home/smco500/.suites/gdps/g2/hub/ppp6/monitoring/banco/'),
    ('evalat', '/home/smco500/.suites/gdps/g2/hub/ppp6/monitoring/banco/'),
    ('dbase', '/home/smco500/.suites/psmon/monitoring/g2/hub/ppp5/sqlite/banco/'),
    ('evalalt', '/home/smco500/.suites/psmon/monitoring/g2/hub/ppp5/sqlite/banco/')
]

family_patterns = ['*cris*', '*cris*']
start_date_str = '2024102500'
end_date_str = '2024102800'

# Get counts for both families
counts = {}
for name, path in families:
    counts[name] = get_counts_by_date(path, family_patterns[0], start_date_str, end_date_str)

# Create DataFrame for easy plotting
df_list = []
for name in counts:
    df = pd.DataFrame(list(counts[name].items()), columns=['Date', f'Count_{name}'])
    df_list.append(df)

# Merge all DataFrames
df_merged = df_list[0]
for df in df_list[1:]:
    df_merged = pd.merge(df_merged, df, on='Date', how='outer')

df_merged.fillna(0, inplace=True)
df_merged.sort_values(by='Date', inplace=True)

# Calculate percentages
df_merged['Percentage_evalalt'] = (df_merged['Count_evalalt'] / df_merged['Count_dbase'].replace(0, pd.NA)) * 100
df_merged['Percentage_postalt'] = (df_merged['Count_postalt'] / df_merged['Count_evalat'].replace(0, pd.NA)) * 100

# Calculate average percentages
average_percentage_evalalt = df_merged['Percentage_evalalt'].mean()
average_percentage_postalt = df_merged['Percentage_postalt'].mean()

# Plotting
fig, ax1 = plt.subplots(figsize=(12, 6))

# Plot percentages on primary y-axis
ax1.plot(df_merged['Date'], df_merged['Percentage_evalalt'], marker='o', color='tab:red', label='Percentage (evalalt / dbase)')
ax1.plot(df_merged['Date'], df_merged['Percentage_postalt'], marker='o', color='tab:orange', label='Percentage (postalt / evalat)')
ax1.set_xlabel('Date')
ax1.set_ylabel('Percentage (%)', color='tab:red')
ax1.tick_params(axis='y', labelcolor='tab:red')
ax1.set_ylim(0, 105)

# Annotate each point with its percentage
for i, (date, percentage_evalalt, percentage_postalt) in enumerate(zip(df_merged['Date'], df_merged[['Percentage_evalalt', 'Percentage_postalt']].values)):
    ax1.text(date, percentage_evalalt, f'{percentage_evalalt:.2f}%', ha='center', va='bottom', fontsize=8, color='tab:red')
    ax1.text(date, percentage_postalt, f'{percentage_postalt:.2f}%', ha='center', va='bottom', fontsize=8, color='tab:orange')

# Secondary y-axis to show total counts for both families
ax2 = ax1.twinx()
ax2.set_ylim(0, df_merged[['Count_evalalt', 'Count_dbase', 'Count_postalt', 'Count_evalat']].max().max() * 1.1)
ax2.plot(df_merged['Date'], df_merged['Count_evalalt'], marker='x', color='tab:blue', label='Count evalalt')
ax2.plot(df_merged['Date'], df_merged['Count_dbase'], marker='x', color='tab:green', label='Count dbase')
ax2.plot(df_merged['Date'], df_merged['Count_postalt'], marker='x', color='tab:purple', label='Count postalt')
ax2.plot(df_merged['Date'], df_merged['Count_evalat'], marker='x', color='tab:brown', label='Count evalat')
ax2.set_ylabel('Observations (Count)', color='black')

# Adding legends
ax1.legend(loc='upper left', bbox_to_anchor=(0, 1.15))
ax2.legend(loc='upper right', bbox_to_anchor=(1, 1.15))

# Display average percentages on the plot
plt.title(f'Percentage Comparison; Average Percentages: evalalt: {average_percentage_evalalt:.2f}%, postalt: {average_percentage_postalt:.2f}%')

# Set grid, rotate x labels, and tight layout
plt.grid(True)
plt.xticks(rotation=45)
plt.tight_layout()
plt.savefig(f'comparison_plot_all_families_{start_date_str}_{end_date_str}.png', dpi=300)

plt.show()




#!/usr/bin/env python3
import numpy as np
import sys
import csv
import dateutil
from datetime import datetime
#import matplotlib.dates as dates2
from matplotlib.dates import date2num
import pikobs
import math
from matplotlib.dates import MONDAY,MO, TU, WE, TH, FR, SA, SU
from matplotlib.dates import  DateFormatter, WeekdayLocator,DayLocator
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as pyplot
import sqlite3
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.dates import DateFormatter, DayLocator, date2num
import dateutil.parser
from matplotlib.lines import Line2D

def custom_round(value):

    
     return round(value, 3)


def plot_time_series(curso, table_name):
              """
              Plot time series data for all distinct `id_stn` in the specified database table.
              
              :param database_path: Path to the SQLite database file.
              :param table_name: Name of the table containing the time series data.
              """
              
              try:
                  # Query to retrieve distinct id_stn values
                  distinct_stations_query = f"SELECT DISTINCT id_stn FROM {table_name};"
                  cursor.execute(distinct_stations_query)
                  distinct_stations = cursor.fetchall()
                  # Extract the station IDs
                  id_stn_values = [row[0] for row in distinct_stations]
          
                  # Define the SQL query to fetch date and Nobs for each station
                  query_template = f"""
                      SELECT
                          DATE,
                          Nacc  -- Make sure this is the correct field for your case
                      FROM
                          {table_name}
                      WHERE
                          id_stn = ?
                      GROUP BY
                          DATE;
                  """
          
                  # Initialize a plot
                  import matplotlib.pyplot as plt
                  plt.figure(figsize=(10, 6))
          
                  # Loop over each id_splot_time_series(curso, table_name)tn value to fetch data and plot
                  for id_stn in id_stn_values:
                      print (id_stn)
                      # Execute the query with the current id_stn
                      cursor.execute(query_template, (id_stn,))
                      results = cursor.fetchall()
          
                      # Extract Dates and Nobs and convert them to arrays
                      Dates = np.array([row[0] for row in results])
                      Nobs = np.array([row[1] for row in results])
                      # Plot the data, label each line with the id_stn
                      plt.plot(Dates, Nobs, label=f'Station {id_stn}')
          
                  # Customize the plot
                  plt.xlabel('Date')
                  plt.ylabel('Nobs')
                  plt.title('Time Series of Nobs for Multiple Stations')
                  plt.legend()
                  plt.xticks(rotation=45)
                  plt.tight_layout()
          
                  # Show the plot
                  pyplot.savefig('po.png')
              finally:

                  print (1)

def timeserie_plot_all_t(pathwork, datestart, dateend, fonction, flag_type, family,
                   region, fig_title, vcoord, id_stn, varnos, files, names):

    import matplotlib as mpl
    import matplotlib.pyplot as plt
    from matplotlib.dates import date2num, DayLocator, HourLocator, DateFormatter
    from matplotlib.ticker import NullFormatter
    from datetime import datetime
    import numpy as np
    import sqlite3
    import dateutil.parser
    import math

    # Ajustes globales de estilo
    mpl.style.use('classic')
    mpl.rcParams.update({
        'lines.linewidth': 1.0,
        'lines.dashed_pattern': [6, 6],
        'lines.dashdot_pattern': [3, 5, 1, 5],
        'lines.dotted_pattern': [1, 3],
        'lines.scale_dashes': False
    })

    if isinstance(varnos, int):
        varnos = [varnos]

    def build_criteria(id_stn, vcoord):
        if id_stn == 'join' and vcoord == 'join':
            return ""
        elif id_stn == 'join':
            return f" and vcoord = {vcoord} "
        elif vcoord == 'join':
            return f" and id_stn = '{id_stn}' "
        else:
            return f" and vcoord = {vcoord} and id_stn = '{id_stn}' "
    from itertools import cycle
    colors_base = plt.get_cmap('tab20').colors
    if varnos!=None:
        varno = varnos
        FNAM, FNAMP, SUM, SUM2 = pikobs.type_boxes(fonction)
        Misgv, MisgvN = -999., 0
        variable_name, units, vcoord_type = pikobs.type_varno(varno)

        try:
            vcoord = int(vcoord)
        except ValueError:
            pass
        for filenumb, file in enumerate(files):
            conn = sqlite3.connect(file)
            cursor = conn.cursor()
            cursor.execute("PRAGMA TEMP_STORE=memory")

            cursor.execute(f"SELECT DISTINCT id_stn FROM timeserie WHERE varno = {varno};")
            id_stns = [row[0] for row in cursor.fetchall()]
            if not id_stns:
                continue

            fig = plt.figure(figsize=(14, 5))
            axes = [
                plt.axes([0.1, 0.58, 0.9, 0.35]),
                plt.axes([0.1, 0.21, 0.9, 0.25]),
                plt.axes([0.1, 0.01, 0.9, 0.15])
            ]
            for i, ax in enumerate(axes):
                ax.grid(True)
                if i == 0:
                    ax.xaxis.tick_bottom()
                else:
                    ax.xaxis.set_visible(i == 1)
                    ax.yaxis.set_visible(i == 1)
            axes[2].xaxis.set_visible(False)
            axes[2].yaxis.set_visible(False)

            xmi, xma = None, None
            TITRE = ""
            legend_items = []

            for idx, id_stn in enumerate(id_stns):
                id_stns_sorted = sorted(set(id_stns))  # todos los id_stn únicos ordenados
                id_stn_color_map = {
                id_stn: colors_base[i % len(colors_base)]
                  for i, id_stn in enumerate(id_stns_sorted)
                    }
                crite = build_criteria(id_stn, vcoord)
                query = f"""
                    SELECT DATE, varno, Nrej, Nacc,
                      SUM({SUM})/SUM(CAST(N AS FLOAT)) AVG,
                      SQRT(SUM({SUM2})/SUM(CAST(N AS FLOAT)) - SUM({SUM})/SUM(CAST(N AS FLOAT))*SUM({SUM})/SUM(CAST(N AS FLOAT))) STD,
                      SUM(sumstat)/SUM(CAST(N AS FLOAT)) BCORR
                    FROM timeserie
                    WHERE varno='{varno}' {crite}
                    GROUP BY DATE;
                """
                cursor.execute(query)
                results = cursor.fetchall()
                if not results:
                    continue

                Dates2 = np.array([row[0] for row in results])
                Nrejets = np.array([row[2] for row in results], dtype=float)
                Bomp2   = np.array([row[4] for row in results], dtype=float)
                Somp2   = np.array([row[5] for row in results], dtype=float)
                Nomb2   = np.array([row[3] for row in results], dtype=float)
                bcorr   = np.array([row[6] for row in results], dtype=float)
                dates = [dateutil.parser.parse(s) for s in Dates2]
                debut, fin = dates[0], dates[-1]
                if not TITRE:
                    TITRE = f'From {debut:%Y%m%d%H} to {fin:%Y%m%d%H}'
                x = date2num(dates)

                Bomp2_masked = np.ma.masked_where(Bomp2 <= Misgv, Bomp2)
                Somp2_masked = np.ma.masked_where(Somp2 <= Misgv, Somp2)
                NB_masked = np.ma.masked_where(Nomb2 <= MisgvN, Nomb2)
                SNrej_masked = np.ma.masked_where(Nrejets <= Misgv, Nrejets)
                Bcorr_masked = np.ma.masked_where(bcorr <= Misgv, bcorr)

                ax1 = axes[0]
                ax2 = axes[1]

               # color = couleurs[idx % len(couleurs)]
               # color = next(couleurs)
                color = id_stn_color_map[id_stn]
                ax1.plot(x, Bomp2_masked, '--o', ms=3, color=color, lw=2)
                ax1.plot(x, Somp2_masked, '-*', ms=4, color=color, lw=2)
                if fonction in ('BIAS_CORR', 'OBS_ERROR'):
                    ax1.plot(x, Bcorr_masked, '-p', ms=4, color='g', lw=2)
                if fonction in ('NOBS', 'DENS'):
                    ax1.plot(x, NB_masked, drawstyle='steps', color='g')

                ax2.plot(x, NB_masked, '-o', drawstyle='steps', ms=3, color=color, lw=2)
                ax2.plot(x, SNrej_masked, '--p', drawstyle='steps', ms=4, color=color, lw=2)

                legend_items.append((id_stn, color))

                if xmi is None:
                    xmi, xma = x[0], x[-1]
                else:
                    xmi = min(xmi, x[0])
                    xma = max(xma, x[-1])

            # Format eje X
            diff_days = (datetime.strptime(dateend, "%Y%m%d%H") - datetime.strptime(datestart, "%Y%m%d%H")).days
            interval = 1 if diff_days <= 30 else 3
            major_locator = DayLocator(interval=interval)
            minor_locator = HourLocator(interval=10)
            major_formatter = DateFormatter('%m%d')

            for ax in [axes[0], axes[1]]:
                ax.set_xlim([xmi, xma])
                ax.xaxis.set_major_locator(major_locator)
                ax.xaxis.set_major_formatter(major_formatter)
                ax.xaxis.set_minor_locator(minor_locator)
                ax.xaxis.set_minor_formatter(NullFormatter())
                ax.tick_params(axis='x', which='minor', bottom=True)
                ax.tick_params(axis='x', which='major', bottom=True, labelsize=10)
                ax.grid(True, which='minor', axis='x')

            axes[0].text(0.00, 1.04, names[filenumb], fontsize=11, color='black', transform=axes[0].transAxes)
            axes[0].text(0.25, 1.04, TITRE, fontsize=11, color='black', transform=axes[0].transAxes)

            axes[2].text(0.01, 0.7, f'Family:{family}', fontsize=12)
            axes[2].text(0.42, 0.7, f'{variable_name} {units}', fontsize=12)
            axes[2].text(0.85, 0.7, f'Region:{region}', fontsize=12)
            
            legendlist = ['μ', 'σ']
            l2 = ax1.legend(legendlist, columnspacing=1, fancybox=True, ncol=2, shadow=False, loc=(0.80, +1.030))
           # [h.set_color('black') for h in l2.legendHandles]

            legendlist2 = ['Nobs', 'Nrej']
            l3 = ax2.legend(legendlist2, columnspacing=1, fancybox=True, ncol=2, shadow=False, loc=(0.80, +1.030))
         #   [h.set_color('black') for h in l3.legendHandles]

            for i, (station, color) in enumerate(legend_items):
                axes[2].text(0.01 + (i % 4) * 0.24, 0.02 + (i // 4) * 0.2,
                             f'{station}', fontsize=11, color=color)
            
            fig.savefig(f'{pathwork}/{family}/timeserie_{region}_{datestart}_{dateend}_{fonction}_allid_stn_varno_{varno}_{names[filenumb]}.png',
                        format='png', dpi=100, bbox_inches='tight')
            plt.close(fig)

def timeserie_plot(
    pathwork, datestart, dateend, fonction, flag_type, family,
    region, fig_title, vcoord, id_stn, varnos, files, names
    ):
    # Ajustes globales mpl
    mpl.style.use('classic')
    mpl.rcParams.update({
    'lines.linewidth': 1.0,
    'lines.dashed_pattern': [6, 6],
    'lines.dashdot_pattern': [3, 5, 1, 5],
    'lines.dotted_pattern': [1, 3],
    'lines.scale_dashes': False
    })
    
    if id_stn=='all_t':
      timeserie_plot_all_t(
    pathwork, datestart, dateend, fonction, flag_type, family,
    region, fig_title, vcoord, id_stn, varnos, files, names
     )
    else:   
       if isinstance(varnos, int):
           varnos = [varnos]
       
       def build_criteria(id_stn, vcoord):
           if id_stn == 'join' and vcoord == 'join':
               return ""
           elif id_stn == 'join':
               return f" and vcoord = {vcoord} "
           elif vcoord == 'join':
               return f" and id_stn = '{id_stn}' "
           else:
               return f" and vcoord = {vcoord} and id_stn = '{id_stn}' "
       
       for varno in varnos:
           FNAM, FNAMP, SUM, SUM2 = pikobs.type_boxes(fonction)
           couleurs = [plt.cm.Set1(i) for i in np.linspace(0, 1, 9)]
           couleurs[0]='blue'
           couleurs[1]='red'
           Misgv, MisgvN = -999., 0
           variable_name, units, vcoord_type = pikobs.type_varno(varno)
           try:
               vcoord = int(vcoord)
           except ValueError:
               pass
       
           famille = f'Family:{family} id_stn:{id_stn} Channel:{vcoord}'
           label = f'Region:{region}'
           vcoord_type_e = f'{variable_name} {units}'
           plt.close('all')
           fig = plt.figure(figsize=(14, 5))
           axes = [
               plt.axes([0.1, 0.58, 0.9, 0.35]),
               plt.axes([0.1, 0.21, 0.9, 0.25]),
               plt.axes([0.1, 0.01, 0.9, 0.15])
           ]
           for i, ax in enumerate(axes):
               ax.grid(True)
               if i == 0:
                   ax.xaxis.tick_bottom()
               else:
                   ax.xaxis.set_visible(i == 1)
                   ax.yaxis.set_visible(i == 1)
           axes[2].xaxis.set_visible(False)
           axes[2].yaxis.set_visible(False)
       
           TITRE = []
           for filenumb, file in enumerate(files):
               conn = sqlite3.connect(file)
               cursor = conn.cursor()
               cursor.execute("PRAGMA TEMP_STORE=memory")
               crite = build_criteria(id_stn, vcoord)
               query = f"""
                   SELECT DATE, varno, Nrej, Nacc,
                     SUM({SUM})/SUM(CAST(N AS FLOAT)) AVG,
                     SQRT(SUM({SUM2})/SUM(CAST(N AS FLOAT)) - SUM({SUM})/SUM(CAST(N AS FLOAT))*SUM({SUM})/SUM(CAST(N AS FLOAT))) STD,
                     SUM(sumstat)/SUM(CAST(N AS FLOAT)) BCORR,
                     count(flag)
                   FROM timeserie
                   WHERE varno='{varno}' {crite}
                   GROUP BY DATE;
               """
               cursor.execute(query)
               results = cursor.fetchall()
               if not results:
                   continue
       
               # Arrays numpy directos
               Dates2 = np.array([row[0] for row in results])
               Nrejets = np.array([row[2] for row in results], dtype=float)
               Bomp2   = np.array([row[4] for row in results], dtype=float)
               Somp2   = np.array([row[5] for row in results], dtype=float)
               Nomb2   = np.array([row[3] for row in results], dtype=float)
               bcorr   = np.array([row[6] for row in results], dtype=float)
               dates = [dateutil.parser.parse(s) for s in Dates2]
               debut, fin = dates[0], dates[-1]
               PERIODE = f' From {debut:%Y%m%d%H} to {fin:%Y%m%d%H}'
               TITRE.append(PERIODE)
               x = date2num(dates)
               # Máscaras compactas
               Bomp2_masked = np.ma.masked_where(Bomp2 <= Misgv, Bomp2)
               Somp2_masked = np.ma.masked_where(Somp2 <= Misgv, Somp2)
               NB_masked = np.ma.masked_where(Nomb2 <= MisgvN, Nomb2)
               SNrej_masked = np.ma.masked_where(Nrejets <= Misgv, Nrejets)
               Bcorr_masked = np.ma.masked_where(bcorr <= Misgv, bcorr)
               Bias, Sigma, Nomb = map(np.ma.compressed, [Bomp2_masked, Somp2_masked, NB_masked])
               xmi, xma = x[0], x[-1]
       
               # Primer plot
               ax1 = axes[0]
               ax1.set_xlim([xmi, xma])
               ax1.plot(x, Bomp2_masked, '--o', ms=3, color=couleurs[filenumb], lw=2)
               ax1.plot(x, Somp2_masked, '-*', ms=4, color=couleurs[filenumb], lw=2)
               if fonction in ('BIAS_CORR', 'OBS_ERROR'):
                   ax1.plot(x, Bcorr_masked, '-p', ms=4, color='g', lw=2)
               if fonction in ('NOBS', 'DENS'):
                   ax1.plot(x, NB_masked, drawstyle='steps', color='g')
               # X-Axis formatos solo una vez
               
               
               from matplotlib.dates import HourLocator
               from matplotlib.ticker import NullFormatter
               from matplotlib.dates import DayLocator, HourLocator, DateFormatter
               from matplotlib.ticker import NullFormatter
               
               # Locatorsdatestart, dateend,
   
               diff_days = (datetime.strptime(datestart, "%Y%m%d%H")-datetime.strptime(dateend, "%Y%m%d%H")).days
   
               interval = 2 if diff_days <= 30 else 4
               major_locator = DayLocator(interval=interval)                
               minor_locator = HourLocator(interval=6)      
               major_formatter = DateFormatter('%m%d')    
               
               for ax in [axes[0], axes[1]]:
                   ax.xaxis.set_major_locator(major_locator)
                   ax.xaxis.set_major_formatter(major_formatter)
               
                   ax.xaxis.set_minor_locator(minor_locator)
                   ax.xaxis.set_minor_formatter(NullFormatter())  
               
                   ax.grid(True, which='minor', axis='x')
               
                   ax.tick_params(axis='x', which='minor', bottom=True)
                   ax.tick_params(axis='x', which='major', bottom=True, labelsize=10)
                
   
               ax2 = axes[1]
               ax2.set_xlim([xmi, xma])
               ax2.plot(x, NB_masked, '-o', drawstyle='steps', ms=3, color=couleurs[filenumb], lw=2)
               ax2.plot(x, SNrej_masked, '--p', drawstyle='steps', ms=4, color=couleurs[filenumb], lw=2)
       
               # Estadísticas y anotaciones
               legendlist = ['μ', 'σ']
               if Nomb.size > 0:
                   Mu = np.sum(Bias * Nomb) / np.sum(Nomb)
                   Sx2 = (Sigma ** 2 + Bias ** 2) * Nomb
                   Sig = math.sqrt(np.sum(Sx2) / np.sum(Nomb) - Mu ** 2)
                   ax1.text(0.45, 1.12 + filenumb / 9., f'{fonction} μ:{custom_round(Mu)} σ:{custom_round(Sig)} Nobs:{int(np.sum(Nomb))}',
                            transform=ax1.transAxes, fontsize=11, va='top', color=couleurs[filenumb])
               # Leyendas y textos, manteniendo resultado
               ax1.text(.00, 1.040, names[0], fontsize=11, color=couleurs[0], transform=ax1.transAxes)
               ax1.text(.14, 1.040, TITRE[0], fontsize=11, color=couleurs[0], transform=ax1.transAxes)
               axes[2].text(.01, 0.24, famille, fontsize=12, color='black', transform=axes[2].transAxes)
               axes[2].text(.42, 0.24, vcoord_type_e, fontsize=12, color='black', transform=axes[2].transAxes)
               axes[2].text(.85, 0.24, label, fontsize=12, color='black', transform=axes[2].transAxes)
               l2 = ax1.legend(legendlist, columnspacing=1, fancybox=True, ncol=2, shadow=False, loc=(0.80, +1.030))
              # [h.set_color('black') for h in l2.legendHandles]
               legendlist2 = ['Nobs', 'Nrej']
               l3 = ax2.legend(legendlist2, columnspacing=1, fancybox=True, ncol=2, shadow=False, loc=(0.80, +1.030))
             #  [h.set_color('black') for h in l3.legendHandles]
               if filenumb == 1 and len(names) > 1:
                   ax1.text(.00, 1.030 + .056 * (filenumb + 1), names[1], fontsize=11, color=couleurs[filenumb], transform=ax1.transAxes)
                   try:
                      ax1.text(.14, 1.030 + .056 * (filenumb + 1), TITRE[1], fontsize=11, color=couleurs[filenumb], transform=ax1.transAxes)
                   except:  
                      ax1.text(.14, 1.030 + .056 * (filenumb + 1), TITRE[0], fontsize=11, color=couleurs[filenumb], transform=ax1.transAxes)
                
                      continue
                #   axes[2].text(.01, 0.016 + .30 * (filenumb + 1), famille, fontsize=12, color=couleurs[filenumb], transform=axes[2].transAxes)
                #   axes[2].text(.32, 0.016 + .30 * (filenumb + 1), vcoord_type_e, fontsize=12, color=couleurs[filenumb], transform=axes[2].transAxes)
                #   axes[2].text(.75, 0.016 + .30 * (filenumb + 1), label, fontsize=12, color=couleurs[filenumb], transform=axes[2].transAxes)
       fig.savefig(f'{pathwork}/{family}/timeserie_{region}_{datestart}_{dateend}_{fonction}_id_stn_{id_stn}_vcoord_{vcoord}_varno_{varno}.png',
                   format='png', dpi=100, bbox_inches='tight')
   




import sqlite3
import matplotlib.pyplot as plt
import numpy as np
from collections import defaultdict
import json
import random
import os

def flags_plot(pathwork, datestart, dateend, family, plot_title, vcoord, id_stn, varno, region):
    db_file = f'{pathwork}/output_file{region}.db'

    def bits_active(number):
        if number == 0:
            return [0]
        return [bit for bit in range(24) if (number & (1 << bit)) != 0]

    results_by_date = defaultdict(lambda: defaultdict(int))
    total_observations_by_date = defaultdict(int)

    # Conexión con la base de datos
    connection = sqlite3.connect(db_file)
    cursor = connection.cursor()

    # Obtener todas las fechas y observaciones
    cursor.execute(f"""
        SELECT DISTINCT date, flag
        FROM flag_observations
        WHERE varno={varno}
    """)
    rows = cursor.fetchall()
    for date, flag in rows:
        cursor.execute(f"""
            SELECT count(*)
            FROM flag_observations
            WHERE date=? AND flag=? AND varno={varno}
        """, (date, flag))
        num_observations = cursor.fetchone()[0]
        results_by_date[date][flag] += num_observations

    cursor.execute(f"""
        SELECT date, count(*)
        FROM flag_observations
        WHERE varno={varno}
        GROUP BY date
    """)
    total_observations = cursor.fetchall()
    for date, count in total_observations:
        total_observations_by_date[date] = count
    
    connection.close()

    if not results_by_date:
        print("No data found.")
        return
    
    normalized_data = defaultdict(lambda: defaultdict(float))
    numberobs_data = defaultdict(lambda: defaultdict(float))
    
    for date, counts in results_by_date.items():
        total_observations = total_observations_by_date[date]
        accumulated = 0
        for flag, num_observations in sorted(counts.items(), key=lambda item: item[1], reverse=True):
            percentage = (num_observations / total_observations) * 100
            normalized_data[date][flag] = (accumulated, accumulated + percentage)
            numberobs_data[date][flag] = num_observations
            accumulated += percentage
    
    average_percentage_by_flag = {}
    average_obs_by_flag = {}
    
    all_flags = set()
    for date in results_by_date:
        all_flags.update(normalized_data[date].keys())
    
    total_observations_all_dates = sum(total_observations_by_date.values())
    for flag in all_flags:
        total_flag_observations = sum(results_by_date[date][flag] for date in results_by_date)
        average_percentage_by_flag[flag] = (total_flag_observations / total_observations_all_dates) * 100
        average_obs_by_flag[flag] = total_flag_observations

    # Color mapping
    mapping_file = 'color_mapping.json'
    
    def load_color_mapping():
        if os.path.exists(mapping_file):
            with open(mapping_file, 'r') as file:
                try:
                    mapping = json.load(file)
                    return mapping
                except json.JSONDecodeError:
                    return {}
        return {}
    
    def save_color_mapping(mapping):
        with open(mapping_file, 'w') as file:
            json.dump(mapping, file, indent=4)
    
    def get_color_for_flag(flag, color_mapping):
        if flag not in color_mapping:
            new_color = generate_random_color()
            color_mapping[flag] = new_color
            save_color_mapping(color_mapping)
        return color_mapping[flag]
    
    def generate_random_color():
        return "#{:06x}".format(random.randint(0, 0xFFFFFF))
    
    color_mapping = load_color_mapping()
    
    all_flags = {8390912, 0, 4, 2240, 8390720}
    for flag in all_flags:
        get_color_for_flag(str(flag), color_mapping)
    
    fig = plt.figure(figsize=(20, 20))
    gs = fig.add_gridspec(14, 5, height_ratios=[3000] + [1000]*13, hspace=1)
    ax1 = fig.add_subplot(gs[0, :])
    fig.subplots_adjust(top=0.9, bottom=0.1, left=0.1, right=0.9, hspace=1, wspace=0.9)
    
    dates = list(total_observations_by_date.keys())
    if len(dates) == 0:
        print("No dates found in the data.")
        return

    bar_width = 1 / len(dates) if len(dates) > 0 else 0.2
    x_positions = []
    for date_idx, date in enumerate(dates):
        sorted_flags = sorted(average_percentage_by_flag.items(), key=lambda item: item[1], reverse=True)
        sorted_flags_keys = [flag for flag, _ in sorted_flags]
        normalized_data[date] = {flag: normalized_data[date][flag] for flag in sorted_flags_keys}
        for flag_idx, flag in enumerate(sorted_flags_keys):
            try:
                start, end = normalized_data[date][flag]
            except KeyError:
                start = 99.999
                end = 100
            
            bits = bits_active(flag)
            avg_percentage = average_percentage_by_flag[flag]
            avg_obs = average_obs_by_flag[flag]
            
            def format_avg_obs(avg_obs, num_dates):
                avg_per_date = avg_obs / num_dates
                if avg_per_date >= 1:
                    return f"{int(round(avg_per_date))}"
                else:
                    return f"{avg_per_date:.1g}"
            
            def format_avg_percentage(avg_percentage):
                if avg_percentage >= 0.001:
                    return f"{avg_percentage:.3f}"
                else:
                    return f"{avg_percentage:.1g}"
            
            formatted_avg_obs = format_avg_obs(avg_obs, len(dates))
            formatted_avg_percentage = format_avg_percentage(avg_percentage)
            label = (f'Flag {flag} (Active bits: {bits}, '
                     f'Avg %: {formatted_avg_percentage}, '
                     f'Avg N. Obs.: {formatted_avg_obs})') if date_idx == 0 else ""
            x_position = date_idx + bar_width * np.arange(len(dates)) - 0.4
            x_positions.extend(x_position)
            ax1.bar(x_position, end - start, bottom=start, width=bar_width, label=label, color=get_color_for_flag(str(flag), color_mapping))
    
    x_min = min(x_positions) if x_positions else 0
    x_max = max(x_positions) + bar_width if x_positions else bar_width
    ax1.set_xlim(x_min, x_max)
    ax1.set_ylim(0, 100)
    ax1.set_yticks(np.arange(0, 101, 20))
    gs.update(wspace=0.3)
    ax1.set_xlabel('Date', fontsize=15)
    ax1.set_ylabel('Percentage of N. Obs.', fontsize=15)
    ax1.set_title(f'Normalized Distribution of Flags and Active Bits per File\n {datestart} to {dateend} {family}', fontsize=15)
    tick_interval = max(1, len(dates) // 6)
    ax1.set_xticks(x_positions[::tick_interval])
    ax1.set_xticklabels(dates[::tick_interval])
    ax1.legend(loc='upper center', bbox_to_anchor=(0.15, -0.4), ncol=1, frameon=False, fontsize=8)
    ax1.tick_params(axis='x', labelsize=18)
    ax1.tick_params(axis='y', labelsize=18)
    
    bit_observations = defaultdict(int)
    total_observations_all_dates = 0
    explanation_text = (
        r"$^1$ Average percentage is calculated based on the total number of observations.\n"
        r"$^2$ Average number of observations per 6-hour intervals."
    )
    fig.text(0.5, 0.5, explanation_text, ha='center', va='top', fontsize=7, linespacing=1.5)
    
    # Segundo gráfico: Observaciones por bit
    fig2 = plt.figure(figsize=(20, 20))
    ax2 = fig2.add_subplot(111)
    
    total_bits = len(bit_observations)
    bit_colors = [get_color_for_flag(str(bit), color_mapping) for bit in bit_observations.keys()]
    sorted_bits = sorted(bit_observations.items())
    bit_labels = [f"Bit {bit}: {count}" for bit, count in sorted_bits]
    
    bar_width = 0.35
    x = np.arange(total_bits)
    
    ax2.bar(x, [count for bit, count in sorted_bits], color=bit_colors, width=bar_width)
    ax2.set_xticks(x)
    ax2.set_xticklabels(bit_labels, rotation=90)
    ax2.set_xlabel('Bits')
    ax2.set_ylabel('Total Observations')
    ax2.set_title('Total Number of Observations per Bit')
    
    plt.show()




def type_varno(varno):
    GRAPHE_NOMVAR = {
        '11215': ('11215:U COMPONENT OF WIND (10M)', '[m/s]', 'review type_varno.py'),
        '11216': ('11216:V COMPONENT OF WIND (10M)', '[m/s]', 'review type_varno.py'),
        '12004': ('12004:DRY BULB TEMPERATURE AT 2M', '', 'review type_varno.py'),
        '10051': ('10051:PRESSURE REDUCED TO MEAN SEA LEVEL', '', 'review type_varno.py'),
        '10004': ('10004:PRESSURE', '', 'review type_varno.py'),
        '12203': ('12203:DEW POINT DEPRESSION (2M)', '', 'review type_varno.py'),
        '12001': ('12001:TEMPERATURE/DRY BULB', '', 'review type_varno.py'),
        '11003': ('11003:U COMPONENT OF WIND', '[m/s]', 'review type_varno.py'),
        '11004': ('11004:V COMPONENT OF WIND', '[m/s]', 'review type_varno.py'),
        '12192': ('12192:DEW POINT DEPRESSION', '', 'review type_varno.py'),
        '12163': ('12163:BRIGHTNESS TEMPERATURE', '[K]', 'Channel'),
        '21014': ('21014:Doppler velocity', '[m/s]', 'Height'),
        '15036': ('15036:ATMOSPHERIC REFRACTIVITY', '', 'review type_varno.py'),
        '11001': ('11001:WIND DIRECTION', '', 'review type_varno.py'),
        '11002': ('11002:WIND SPEED', '', 'review type_varno.py'),
        '11011': ('11011:WIND DIRECTION AT 10M', '', 'review type_varno.py'),
        '11012': ('11012:WIND SPEED AT 10M', '[m/s]', 'review type_varno.py'),
        '11214': ('11214:VECTOR DIFFERENCE OF WIND', '', 'review type_varno.py'),
        '15198': ('15198:TOTAL OR PARTIAL COLUMN (DOBSON)', '', 'review type_varno.py'),
        '15008': ('15008:VOLUMETRIC MIXING RATIO(PROPORTIONED)', '', 'review type_varno.py')
    }

    # Ensure varno is a string for dictionary lookup
    varno_str = str(varno)

    # Check if the variable number is in the dictionary
    if varno_str in GRAPHE_NOMVAR:
        # If it is, retrieve the variable name and units
        variable_name, units, vcoord_type = GRAPHE_NOMVAR[varno_str]
        # Return the variable name along with its units and vcoord_type
        return variable_name, units, vcoord_type
    else:
        # If the variable number is not found, raise an exception
        raise ValueError(f"Varno '{varno}' not found in the dictionary")

